using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using System.IO;

namespace Dark
{
    public enum NodeType
    {
        Unknown,
        Reflexive,
        Tag,
        ID,
        StringID,
        Script,
        Unused,
        Undefined,
        Float,
        Int,
        UInt,
        Short,
        UShort,
        Byte,
        Bitmask,
        Enum,
        Option,
        String,
        Unicode
    }

    public class BaseNode
    {
        public Map map;
        public StructNode Parent;
        public string Name;
        public int Offset;
        public bool Visible;
        public int Size;
        public NodeType NodeType;

        public bool IsVisible;

        public virtual void UpdateVisibility() { }
        public virtual void InitializeControls() { }

        public virtual void ReadValue(BinaryReader br, int seekOffset) { }
        public virtual void SaveChanges(BinaryWriter bw, int seekOffset) { }
        public virtual int DisplayControls(Panel pnl, int xpos, int ypos) { return ypos; }
    }

    public class StructNode : BaseNode
    {
        //Specific
        public int yposPanel;
        public Color Colour;
        public bool Enabled;

        //Controls
        public Panel panel;
        public Label label;
        public ComboBox combobox;

        //Values
        public int chunkCount, translation, atmChunk;

        public StructNode(Map map, StructNode Parent, string Name, string Offset, string Visible, string Size)
        {
            this.map = map;
            this.Parent = Parent;
            this.Name = Name;
            this.Offset = int.Parse(Offset);
            this.Visible = bool.Parse(Visible);
            this.Size = int.Parse(Size);
            this.NodeType = NodeType.Reflexive;
        }

        public override void UpdateVisibility()
        {
            IsVisible = Globals.MetaEditorShownItems[0];
        }

        public override void InitializeControls()
        {
            panel = new Panel();
            label = new Label();
            combobox = new ComboBox();
        }

        public override void ReadValue(BinaryReader br, int seekOffset)
        {
            br.BaseStream.Position = (seekOffset == -1 ?
                map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;

            chunkCount = br.ReadInt32();
            translation = br.ReadInt32();

            translation -= map.tags.tagsList[map.tags.selectedIndex].Type == "sbsp" ?
                map.sbsp.sbspInfoList[map.sbsp.FindIndexByTagID(map.tags.tagsList[map.tags.selectedIndex].Ident)].Magic
                : map.index.secondaryMagic;

            atmChunk = translation;
            if (combobox.SelectedIndex != -1) atmChunk += (combobox.SelectedIndex * this.Size);

            if (translation < 0) Enabled = panel.Enabled = false;
            else
            {
                Enabled = panel.Enabled = true;
                combobox.Items.Clear();
                combobox.Items.Add("0"); //wil always hav at least 1 chunk
                combobox.SelectedIndex = 0;
                combobox.Tag = null;
            }
        }

        public override int DisplayControls(Panel pnl, int xpos, int ypos)
        {
            //panel colour
            Color clr = Parent == null ? pnl.BackColor : Parent.Colour;

            int tmpR = clr.R + Globals.rShift;
            tmpR = tmpR > 255 ? Globals.ReflexiveStartingColor.R : tmpR;

            int tmpG = clr.G + Globals.gShift;
            tmpG = tmpG > 255 ? Globals.ReflexiveStartingColor.G : tmpG;

            int tmpB = clr.B + Globals.bShift;
            tmpB = tmpB > 255 ? Globals.ReflexiveStartingColor.B : tmpB;

            this.Colour = Color.FromArgb(tmpR, tmpG, tmpB);

            //make panel
            panel.Controls.Clear();
            panel.Left = xpos;
            panel.Top = ypos;
            panel.Width = pnl.Width - 24; // (xpos * 2)
            panel.BackColor = Colour;
            panel.BorderStyle = BorderStyle.Fixed3D;

            //relevant to panel
            yposPanel = 6;

            //make label
            label.AutoSize = true;
            label.Text = this.Name;
            label.Left = xpos;
            label.Top = yposPanel;
            panel.Controls.Add(label);
            yposPanel += label.Height + 3;

            //make combobox
            combobox.Left = xpos;
            combobox.Top = yposPanel;
            combobox.Width = panel.Width - 24; // (xpos * 2)

            //cost efficient
            if (this.Enabled)
            {
                panel.MouseEnter += new EventHandler(panel_MouseEnter);
                panel.MouseLeave += new EventHandler(panel_MouseLeave);

                combobox.DropDownStyle = ComboBoxStyle.DropDownList;
                combobox.DropDown += new EventHandler(combobox_DropDown);
                combobox.SelectedIndexChanged += new EventHandler(combobox_SelectedIndexChanged);
            }

            panel.Controls.Add(combobox);
            combobox.Anchor = AnchorStyles.Top | AnchorStyles.Left | AnchorStyles.Right;

            yposPanel += combobox.Height + 12;
            panel.Height = yposPanel;

            pnl.Controls.Add(panel);

            if (Parent != null) panel.Anchor = AnchorStyles.Top | AnchorStyles.Left | AnchorStyles.Right;
            else map.entReader.RControls.Add(panel);

            return ypos + panel.Height + 6;
        }

        void panel_MouseLeave(object sender, EventArgs e)
        {
            map.ProgressLabel.Text = "Idle";
            //map.MetaEditorDescriptor.Height = map.MetaEditorDescriptor.PreferredHeight;
        }

        void panel_MouseEnter(object sender, EventArgs e)
        {
            map.ProgressLabel.Text = 
                "Idle"; //test
        }

        void combobox_DropDown(object sender, EventArgs e)
        {
            if (combobox.Tag == null)
            {
                for (int x = 1; x < chunkCount; x++) combobox.Items.Add(x.ToString());
                combobox.Tag = "1";
            }
        }

        void combobox_SelectedIndexChanged(object sender, EventArgs e)
        {
            BinaryReader br = new BinaryReader(new FileStream(map.path, FileMode.Open,
              FileAccess.ReadWrite, FileShare.ReadWrite));

            atmChunk = translation + (combobox.SelectedIndex * this.Size);

            //loop through nodes
            for (int x = 0; x < map.entReader.Nodes.Count; x++)
            {
                //has a parent (is within a panel)
                if (map.entReader.Nodes[x].IsVisible && map.entReader.Nodes[x].Parent != null)
                {
                    //the current node
                    BaseNode bn = map.entReader.Nodes[x];

                    //will loop at least once
                    while (bn.Parent != null)
                    {
                        if (bn.Parent == this)
                        {
                            //if true, this reflexives translation needs to be fixed, so pass the chunkOffset
                            if (map.entReader.Nodes[x].NodeType == NodeType.Reflexive) map.entReader.Nodes[x].ReadValue(br, atmChunk);
                            //pass on the translated reflexive
                            else if (map.entReader.Nodes[x].Parent.Enabled) map.entReader.Nodes[x].ReadValue(br, map.entReader.Nodes[x].Parent.atmChunk);

                            break;
                        }
                        else bn = bn.Parent;
                    }
                }
            }

            br.Close();
        }
    }

    public class TagNode : BaseNode
    {
        //Specific
        //public IDNode IDnode; //not needed

        //Controls
        Label label;
        ComboBox combobox;
        ComboBox combobox2;
        Label label2;

        //Values
        public string tagType;
        public bool tagTypeInMapIsNull;
        public int ident;
        public string tagPath;

        public TagNode(Map map, StructNode Parent, string Name, string Offset, string Visible)
        {
            this.map = map;
            this.Parent = Parent;
            this.Name = Name;
            this.Offset = int.Parse(Offset);
            this.Visible = bool.Parse(Visible);
            this.Size = 4;
            this.NodeType = NodeType.Tag;
        }

        public override void InitializeControls()
        {


            label = new Label();
            combobox = new ComboBox();
            combobox2 = new ComboBox();
            label2 = new Label();


        }

        public override void UpdateVisibility()
        {
            IsVisible = Globals.MetaEditorShownItems[1];
        }

        public override void ReadValue(BinaryReader br, int seekOffset)
        {


            br.BaseStream.Position = (seekOffset == -1 ?
                map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;

            //read 4 bytes
            byte[] temptagtype = br.ReadBytes(4);
            //reverse
            Array.Reverse(temptagtype);
            //clear
            tagType = "";
            //convert bytes to string
            for (int x = 0; x < 4; x++) tagType += Convert.ToChar(temptagtype[x]);

            //read ID
            ident = br.ReadInt32();
            if (ident == -1) tagPath = "Nulled Out";
            else
            {
                int indx = map.tags.FindTagIndexByID(ident);
                tagPath = indx == -1 ? "Nulled Out" : map.tags.tagsList[indx].Path;
            }

            //if true, hasnt been called before 'DisplayVisually()'
            if (combobox.Items.Count != 0)
            {
                //combobox does contain types, but wrong one is selected
                if (combobox.SelectedItem.ToString() != tagType)
                {
                    //store
                    string tmp = tagPath;

                    //which will add and make tagpath 'nulled out'   
                    if ((combobox.SelectedIndex = combobox.FindString(tagType)) == -1)
                    {
                        combobox.SelectedIndex = 0;
                        tagTypeInMapIsNull = true;
                    }
                    else tagTypeInMapIsNull = false;

                    combobox2.Items.Clear();
                    combobox2.SelectedIndex = combobox2.Items.Add(tmp);
                }
                else //combobox does contain types, and right one is selected
                {
                    //only contains 'nulled out'
                    if (combobox2.Tag == null)
                    {
                        combobox2.Items.Clear();
                        combobox2.SelectedIndex = combobox2.Items.Add(tagPath);
                    }
                    else combobox2.SelectedIndex = combobox2.FindString(tagPath);
                }
            }


        }

        public override void SaveChanges(BinaryWriter bw, int seekOffset)
        {
            bw.BaseStream.Position = (seekOffset == -1 ?
                map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;

            if (combobox.SelectedIndex == 0)
            {
                if (!tagTypeInMapIsNull)
                {
                    bw.Write("".ToCharArray()); //FFFF FFFF
                    tagTypeInMapIsNull = true;
                }
            }
            else
            {
                bw.Write(map.Reverse(tagType).ToCharArray());
                tagTypeInMapIsNull = false;
            }
            bw.Write(ident);
        }

        public override int DisplayControls(Panel pnl, int xpos, int ypos)
        {


            //label
            label.AutoSize = true;
            label.Text = "Tag";
            label.Left = xpos;
            label.Top = ypos;
            pnl.Controls.Add(label);

            xpos += label.Width + 2;
            ypos -= 3;

            //combobox of types
            combobox.Left = xpos;
            combobox.Top = ypos;
            combobox.Width = 90;
            pnl.Controls.Add(combobox);

            xpos += combobox.Width + 2;

            //combobox for paths
            combobox2.Left = xpos;
            combobox2.Top = ypos;
            combobox2.Width = 255;
            pnl.Controls.Add(combobox2);

            xpos += combobox2.Width + 2;
            ypos += 3;

            //label
            label2.AutoSize = true;
            label2.Text = this.Name;
            label2.Left = xpos;
            label2.Top = ypos;
            pnl.Controls.Add(label2);

            //cost efficient
            if (Parent == null || Parent.Enabled)
            {
                //combobox
                combobox.DropDownStyle = ComboBoxStyle.DropDownList;
                combobox.Items.Add("Nulled Out");
                combobox.Items.AddRange(map.tags.tagTypes.ToArray());
                if ((combobox.SelectedIndex = combobox.FindString(tagType)) == -1)
                {
                    combobox.SelectedIndex = 0;
                    tagTypeInMapIsNull = true;
                }
                else tagTypeInMapIsNull = false;
                combobox.SelectedIndexChanged += new EventHandler(combobox_SelectedIndexChanged);

                //combobox for paths
                combobox2.DropDownStyle = ComboBoxStyle.DropDownList;
                combobox2.Items.Clear();
                combobox2.SelectedIndex = combobox2.Items.Add(tagPath);
                combobox2.DropDown += new EventHandler(combobox2_DropDown);
                combobox2.SelectedIndexChanged += new EventHandler(combobox2_SelectedIndexChanged);
            }

            return ypos + combobox.Height + 6;
        }

        void combobox_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (combobox.SelectedItem != null)
                tagType = combobox.SelectedItem.ToString();

            combobox2.Items.Clear();
            combobox2.Items.Add("Nulled Out");
            combobox2.SelectedIndex = 0;
            combobox2.Tag = null;
        }

        void combobox2_DropDown(object sender, EventArgs e)
        {
            if (combobox2.Tag == null && combobox.SelectedIndex != 0)
            {
                string sItem = combobox2.SelectedItem.ToString();
                combobox2.Items.Clear();
                combobox2.Items.Add("Nulled Out");
                combobox2.Items.AddRange(map.tags.LoadTagsFromTypeIntoStrings(tagType));
                combobox2.SelectedIndex = combobox2.FindString(sItem);
                combobox2.Tag = "1";
            }
        }

        void combobox2_SelectedIndexChanged(object sender, EventArgs e)
        {
            tagPath = combobox2.SelectedItem.ToString();

            ident = combobox2.SelectedIndex == 0 ? -1 : map.tags.tagsList[
                map.tags.FindTagIndexByTypeAndPath(tagType, tagPath)].Ident;

        }
    }

    public class IDNode : BaseNode
    {
        //Controls
        Label label;
        ComboBox combobox;
        ComboBox combobox2;
        Label label2;

        //Values
        public string tagType;
        public int ident;
        public string tagPath;

        public IDNode(Map map, StructNode Parent, string Name, string Offset, string Visible)
        {
            this.map = map;
            this.Parent = Parent;
            this.Name = Name;
            this.Offset = int.Parse(Offset);
            this.Visible = bool.Parse(Visible);
            this.Size = 4;
            this.NodeType = NodeType.ID;
        }

        public override void InitializeControls()
        {
            this.

            label = new Label();
            combobox = new ComboBox();
            combobox2 = new ComboBox();
            label2 = new Label();


        }

        public override void UpdateVisibility()
        {
            IsVisible = Globals.MetaEditorShownItems[2];
        }

        public override void ReadValue(BinaryReader br, int seekOffset)
        {


            br.BaseStream.Position = (seekOffset == -1 ?
                map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;

            ident = br.ReadInt32();
            int indx = map.tags.FindTagIndexByID(ident);
            tagType = map.tags.tagsList[indx].Type;
            tagPath = map.tags.tagsList[indx].Path;

            //hasnt been called before 'DisplayVisually()'
            if (combobox.Items.Count != 0)
            {
                //combobox does contain types, but wrong one is selected
                if (combobox.SelectedItem != null && combobox.SelectedItem.ToString() != tagType)
                {
                    string tmp = tagPath;
                    //which will add all tagpaths to combobox2
                    combobox.SelectedIndex = combobox.FindString(tagType);

                    combobox2.SelectedIndex = combobox2.FindString(tmp);
                }
                //combobox does contain types, and right one is selected
                else
                {
                    //only contains one path
                    if (combobox2.Tag == null)
                    {
                        combobox2.SelectedIndex = combobox2.Items.Add(tagPath);
                    }
                    else combobox2.SelectedIndex = combobox2.FindString(tagPath);
                }
            }


        }

        public override void SaveChanges(BinaryWriter bw, int seekOffset)
        {
            bw.BaseStream.Position = (seekOffset == -1 ?
                map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;

            bw.Write(ident);
        }

        public override int DisplayControls(Panel pnl, int xpos, int ypos)
        {


            //label
            label.AutoSize = true;
            label.Text = "Lone ID";
            label.Left = xpos;
            label.Top = ypos;
            pnl.Controls.Add(label);

            xpos += label.Width + 2;
            ypos -= 3;

            //combobox of types
            combobox.Left = xpos;
            combobox.Top = ypos;
            combobox.Width = 90;
            pnl.Controls.Add(combobox);

            xpos += combobox.Width + 2;

            //combobox for tags
            combobox2.Left = xpos;
            combobox2.Top = ypos;
            combobox2.Width = 255;
            pnl.Controls.Add(combobox2);

            xpos += combobox2.Width + 2;
            ypos += 3;

            //label
            label2.AutoSize = true;
            label2.Text = this.Name;
            label2.Left = xpos;
            label2.Top = ypos;
            pnl.Controls.Add(label2);

            //cost efficient
            if (Parent == null || Parent.Enabled)
            {
                //combobox
                combobox.DropDownStyle = ComboBoxStyle.DropDownList;
                combobox.Items.AddRange(map.tags.tagTypes.ToArray());
                combobox.SelectedIndex = combobox.FindString(tagType);
                combobox.SelectedIndexChanged += new EventHandler(combobox_SelectedIndexChanged);

                //combobox for paths
                combobox2.DropDownStyle = ComboBoxStyle.DropDownList;
                combobox2.SelectedIndex = combobox2.Items.Add(tagPath);
                combobox2.DropDown += new EventHandler(combobox2_DropDown);
                combobox2.SelectedIndexChanged += new EventHandler(combobox2_SelectedIndexChanged);
            }

            return ypos + combobox.Height + 6;
        }

        void combobox_SelectedIndexChanged(object sender, EventArgs e)
        {
            tagType = combobox.SelectedItem.ToString();

            combobox2.Items.Clear();
            combobox2.Items.AddRange(map.tags.LoadTagsFromTypeIntoStrings(tagType));
            combobox2.SelectedIndex = 0;
            combobox2.Tag = "1";
        }

        void combobox2_DropDown(object sender, EventArgs e)
        {
            if (combobox2.Tag == null)
            {
                string sItem = combobox2.SelectedItem.ToString();
                combobox2.Items.Clear();
                combobox2.Items.AddRange(map.tags.LoadTagsFromTypeIntoStrings(tagType));
                combobox2.SelectedIndex = combobox2.FindString(sItem);
                combobox2.Tag = "1";
            }
        }

        void combobox2_SelectedIndexChanged(object sender, EventArgs e)
        {
            tagPath = combobox2.SelectedItem.ToString();

            ident = combobox2.SelectedIndex == 0 ? -1 : map.tags.tagsList[
                map.tags.FindTagIndexByTypeAndPath(tagType, tagPath)].Ident;
        }
    }

    public class StringIDNode : BaseNode
    {
        //Controls
        Label label;
        ComboBox combobox;
        Label label2;

        //Values
        public short strIndex;

        public StringIDNode(Map map, StructNode Parent, string Name, string Offset, string Visible)
        {
            this.map = map;
            this.Parent = Parent;
            this.Name = Name;
            this.Offset = int.Parse(Offset);
            this.Visible = bool.Parse(Visible);
            this.Size = 4;
            this.NodeType = NodeType.StringID;
        }

        public override void InitializeControls()
        {


            label = new Label();
            combobox = new ComboBox();
            label2 = new Label();


        }

        public override void UpdateVisibility()
        {
            IsVisible = Globals.MetaEditorShownItems[3];
        }

        public override void ReadValue(BinaryReader br, int seekOffset)
        {


            br.BaseStream.Position = (seekOffset == -1 ?
               map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;

            strIndex = br.ReadInt16();

            if (combobox.Tag == null)
            {
                combobox.Items.Clear();
                if (strIndex >= 0 && strIndex < map.strings.stringsList.Length)
                {
                    combobox.Items.Add(map.strings.stringsList[strIndex]);
                    combobox.SelectedIndex = 0;
                }
            }
            else combobox.SelectedIndex = strIndex;


        }

        public override void SaveChanges(BinaryWriter bw, int seekOffset)
        {
            bw.BaseStream.Position = (seekOffset == -1 ?
                map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;

            bw.Write(strIndex);
        }

        public override int DisplayControls(Panel pnl, int xpos, int ypos)
        {


            label.AutoSize = true;
            label.Text = "S-ID";
            label.Left = xpos;
            label.Top = ypos;
            pnl.Controls.Add(label);
            xpos += label.Width + 2;

            ypos -= 3;
            combobox.Left = xpos;
            combobox.Top = ypos;
            combobox.Width = 255;
            pnl.Controls.Add(combobox);
            xpos += combobox.Width + 2;

            ypos += 3;
            label2.AutoSize = true;
            label2.Text = this.Name;
            label2.Left = xpos;
            label2.Top = ypos;
            pnl.Controls.Add(label2);

            //cost efficient
            if (Parent == null || Parent.Enabled)
            {
                combobox.DropDownStyle = ComboBoxStyle.DropDownList;
                combobox.DropDown += new EventHandler(combobox_DropDown);
                combobox.SelectedIndexChanged += new EventHandler(combobox_SelectedIndexChanged);
            }

            return ypos + combobox.Height + 6;
        }

        void combobox_DropDown(object sender, EventArgs e)
        {
            if (combobox.Tag == null)
            {
                combobox.Items.Clear();
                combobox.Items.AddRange(map.strings.stringsList);
                combobox.SelectedIndex = strIndex;
                combobox.Tag = "1";
            }
        }

        void combobox_SelectedIndexChanged(object sender, EventArgs e)
        {
            strIndex = (short)combobox.SelectedIndex;
        }
    }

    public class UnusedNode : BaseNode
    {
        //Controls
        Label label;
        TextBox textbox;
        Label label2;

        public UnusedNode(Map map, StructNode Parent, string Name, string Offset, string Visible, string Size)
        {
            this.map = map;
            this.Parent = Parent;
            this.Name = Name;
            this.Offset = int.Parse(Offset);
            this.Visible = Visible == null ? false : bool.Parse(Visible);
            this.Size = int.Parse(Size);
            this.NodeType = NodeType.Unused;
        }

        public override void InitializeControls()
        {


            label = new Label();
            textbox = new TextBox();
            label2 = new Label();


        }

        public override void UpdateVisibility()
        {
            IsVisible = Globals.MetaEditorShownItems[5];
        }

        public override int DisplayControls(Panel pnl, int xpos, int ypos)
        {
            label.AutoSize = true;
            label.Text = "Unused";
            label.Left = xpos;
            label.Top = ypos;
            pnl.Controls.Add(label);
            xpos += label.Width + 2;

            ypos -= 3;
            textbox.Left = xpos;
            textbox.Top = ypos;
            textbox.Enabled = false;
            pnl.Controls.Add(textbox);
            xpos += textbox.Width + 2;

            ypos += 3;
            label2.AutoSize = true;
            label2.Text = this.Size.ToString() + " bytes";
            label2.Left = xpos;
            label2.Top = ypos;
            pnl.Controls.Add(label2);

            return ypos + textbox.Height + 6;
        }
    }

    public class UndefinedNode : BaseNode
    {
        //Controls
        Label label;
        TextBox textbox;
        Label label2;

        //Values
        float val = 0;

        public UndefinedNode(Map map, StructNode Parent, string Name, string Offset, string Visible)
        {
            this.map = map;
            this.Parent = Parent;
            this.Name = Name;
            this.Offset = int.Parse(Offset);
            this.Visible = bool.Parse(Visible);
            this.Size = 4;
            this.NodeType = NodeType.Undefined;
        }

        public override void InitializeControls()
        {
            label = new Label();
            textbox = new TextBox();
            label2 = new Label();
        }

        public override void UpdateVisibility()
        {
            IsVisible = Globals.MetaEditorShownItems[6];
        }

        public override void ReadValue(BinaryReader br, int seekOffset)
        {
            br.BaseStream.Position = (seekOffset == -1 ? map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;
            val = br.ReadSingle();
            textbox.Tag = val.ToString();
            textbox.Text = val.ToString();
        }

        public override void SaveChanges(BinaryWriter bw, int seekOffset)
        {
            bw.BaseStream.Position = (seekOffset == -1 ?
                map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;

            bw.Write(val);
        }

        public override int DisplayControls(Panel pnl, int xpos, int ypos)
        {
            label.AutoSize = true;
            label.Text = "Undefined";
            label.Left = xpos;
            label.Top = ypos;
            pnl.Controls.Add(label);
            xpos += label.Width + 2;

            ypos -= 3;
            textbox.Left = xpos;
            textbox.Top = ypos;
            pnl.Controls.Add(textbox);
            xpos += textbox.Width + 2;

            ypos += 3;
            label2.AutoSize = true;
            label2.Text = this.Name;
            label2.Left = xpos;
            label2.Top = ypos;
            pnl.Controls.Add(label2);

            //cost efficient
            if (Parent == null || Parent.Enabled)
            {
                textbox.TextChanged += new EventHandler(textbox_TextChanged);
            }

            return ypos + textbox.Height + 6;
        }

        void textbox_TextChanged(object sender, EventArgs e)
        {
            float tmp;
            if (float.TryParse(textbox.Text, out tmp))
            {
                val = tmp;
                textbox.Tag = textbox.Text;
            }
            else textbox.Text = textbox.Tag.ToString();
        }
    }

    public class FloatNode : BaseNode
    {
        //Controls
        Label label;
        TextBox textbox;
        Label label2;

        //Values
        float val = 0;

        public FloatNode(Map map, StructNode Parent, string Name, string Offset, string Visible)
        {
            this.map = map;
            this.Parent = Parent;
            this.Name = Name;
            this.Offset = int.Parse(Offset);
            this.Visible = bool.Parse(Visible);
            this.Size = 4;
            this.NodeType = NodeType.Float;
        }

        public override void InitializeControls()
        {


            label = new Label();
            textbox = new TextBox();
            label2 = new Label();


        }

        public override void UpdateVisibility()
        {
            IsVisible = Globals.MetaEditorShownItems[7];
        }

        public override void ReadValue(BinaryReader br, int seekOffset)
        {


            br.BaseStream.Position = (seekOffset == -1 ? map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;
            val = br.ReadSingle();
            textbox.Tag = val.ToString();
            textbox.Text = val.ToString();


        }

        public override void SaveChanges(BinaryWriter bw, int seekOffset)
        {
            bw.BaseStream.Position = (seekOffset == -1 ?
                map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;

            bw.Write(val);
        }

        public override int DisplayControls(Panel pnl, int xpos, int ypos)
        {


            label.AutoSize = true;
            label.Text = "Float";
            label.Left = xpos;
            label.Top = ypos;
            pnl.Controls.Add(label);
            xpos += label.Width + 2;

            ypos -= 3;
            textbox.Left = xpos;
            textbox.Top = ypos;
            pnl.Controls.Add(textbox);
            xpos += textbox.Width + 2;

            ypos += 3;
            label2.AutoSize = true;
            label2.Text = this.Name;
            label2.Left = xpos;
            label2.Top = ypos;
            pnl.Controls.Add(label2);

            //cost efficient
            if (Parent == null || Parent.Enabled)
            {
                textbox.TextChanged += new EventHandler(textbox_TextChanged);
            }

            return ypos + textbox.Height + 6;
        }

        void textbox_TextChanged(object sender, EventArgs e)
        {
            float tmp;
            if (float.TryParse(textbox.Text, out tmp))
            {
                val = tmp;
                textbox.Tag = textbox.Text;
            }
            else textbox.Text = textbox.Tag.ToString();
        }
    }

    public class Int32Node : BaseNode
    {
        //Controls
        Label label;
        TextBox textbox;
        Label label2;

        //Values
        int val = 0;

        public Int32Node(Map map, StructNode Parent, string Name, string Offset, string Visible)
        {
            this.map = map;
            this.Parent = Parent;
            this.Name = Name;
            this.Offset = int.Parse(Offset);
            this.Visible = bool.Parse(Visible);
            this.Size = 4;
            this.NodeType = NodeType.Int;
        }

        public override void InitializeControls()
        {


            label = new Label();
            textbox = new TextBox();
            label2 = new Label();


        }

        public override void UpdateVisibility()
        {
            IsVisible = Globals.MetaEditorShownItems[8];
        }

        public override void ReadValue(BinaryReader br, int seekOffset)
        {


            br.BaseStream.Position = (seekOffset == -1 ? map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;
            val = br.ReadInt32();
            textbox.Tag = val.ToString();
            textbox.Text = val.ToString();


        }

        public override void SaveChanges(BinaryWriter bw, int seekOffset)
        {
            bw.BaseStream.Position = (seekOffset == -1 ?
                map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;

            bw.Write(val);
        }

        public override int DisplayControls(Panel pnl, int xpos, int ypos)
        {


            label.AutoSize = true;
            label.Text = "Int32";
            label.Left = xpos;
            label.Top = ypos;
            pnl.Controls.Add(label);
            xpos += label.Width + 2;

            ypos -= 3;
            textbox.Left = xpos;
            textbox.Top = ypos;
            pnl.Controls.Add(textbox);
            xpos += textbox.Width + 2;

            ypos += 3;
            label2.AutoSize = true;
            label2.Text = this.Name;
            label2.Left = xpos;
            label2.Top = ypos;
            pnl.Controls.Add(label2);

            //cost efficient
            if (Parent == null || Parent.Enabled)
            {
                textbox.TextChanged += new EventHandler(textbox_TextChanged);
            }

            return ypos + textbox.Height + 6;
        }

        void textbox_TextChanged(object sender, EventArgs e)
        {
            int tmp;
            if (int.TryParse(textbox.Text, out tmp))
            {
                val = tmp;
                textbox.Tag = textbox.Text;
            }
            else textbox.Text = textbox.Tag.ToString();
        }
    }

    public class UInt32Node : BaseNode
    {
        //Controls
        Label label;
        TextBox textbox;
        Label label2;

        //Values
        uint val = 0;

        public UInt32Node(Map map, StructNode Parent, string Name, string Offset, string Visible)
        {
            this.map = map;
            this.Parent = Parent;
            this.Name = Name;
            this.Offset = int.Parse(Offset);
            this.Visible = bool.Parse(Visible);
            this.Size = 4;
            this.NodeType = NodeType.UInt;
        }

        public override void InitializeControls()
        {


            label = new Label();
            textbox = new TextBox();
            label2 = new Label();


        }

        public override void UpdateVisibility()
        {
            IsVisible = Globals.MetaEditorShownItems[9];
        }

        public override void ReadValue(BinaryReader br, int seekOffset)
        {


            br.BaseStream.Position = (seekOffset == -1 ? map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;
            val = br.ReadUInt32();
            textbox.Tag = val.ToString();
            textbox.Text = val.ToString();


        }

        public override void SaveChanges(BinaryWriter bw, int seekOffset)
        {
            bw.BaseStream.Position = (seekOffset == -1 ?
                map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;

            bw.Write(val);
        }

        public override int DisplayControls(Panel pnl, int xpos, int ypos)
        {


            label.AutoSize = true;
            label.Text = "UInt32";
            label.Left = xpos;
            label.Top = ypos;
            pnl.Controls.Add(label);
            xpos += label.Width + 2;

            ypos -= 3;
            textbox.Left = xpos;
            textbox.Top = ypos;
            pnl.Controls.Add(textbox);
            xpos += textbox.Width + 2;

            ypos += 3;
            label2.AutoSize = true;
            label2.Text = this.Name;
            label2.Left = xpos;
            label2.Top = ypos;
            pnl.Controls.Add(label2);

            //cost efficient
            if (Parent == null || Parent.Enabled)
            {
                textbox.TextChanged += new EventHandler(textbox_TextChanged);
            }

            return ypos + textbox.Height + 6;
        }

        void textbox_TextChanged(object sender, EventArgs e)
        {
            uint tmp;
            if (uint.TryParse(textbox.Text, out tmp))
            {
                val = tmp;
                textbox.Tag = textbox.Text;
            }
            else textbox.Text = textbox.Tag.ToString();
        }
    }

    public class ShortNode : BaseNode
    {
        //Controls
        Label label;
        TextBox textbox;
        Label label2;

        //Values
        short val = 0;

        public ShortNode(Map map, StructNode Parent, string Name, string Offset, string Visible)
        {
            this.map = map;
            this.Parent = Parent;
            this.Name = Name;
            this.Offset = int.Parse(Offset);
            this.Visible = bool.Parse(Visible);
            this.Size = 2;
            this.NodeType = NodeType.Short;
        }

        public override void InitializeControls()
        {


            label = new Label();
            textbox = new TextBox();
            label2 = new Label();


        }

        public override void UpdateVisibility()
        {
            IsVisible = Globals.MetaEditorShownItems[10];
        }

        public override void ReadValue(BinaryReader br, int seekOffset)
        {


            br.BaseStream.Position = (seekOffset == -1 ? map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;
            val = br.ReadInt16();
            textbox.Tag = val.ToString();
            textbox.Text = val.ToString();


        }

        public override void SaveChanges(BinaryWriter bw, int seekOffset)
        {
            bw.BaseStream.Position = (seekOffset == -1 ?
                map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;

            bw.Write(val);
        }

        public override int DisplayControls(Panel pnl, int xpos, int ypos)
        {


            label.AutoSize = true;
            label.Text = "Short";
            label.Left = xpos;
            label.Top = ypos;
            pnl.Controls.Add(label);
            xpos += label.Width + 2;

            ypos -= 3;
            textbox.Left = xpos;
            textbox.Top = ypos;
            pnl.Controls.Add(textbox);
            xpos += textbox.Width + 2;

            ypos += 3;
            label2.AutoSize = true;
            label2.Text = this.Name;
            label2.Left = xpos;
            label2.Top = ypos;
            pnl.Controls.Add(label2);

            //cost efficient
            if (Parent == null || Parent.Enabled)
            {
                textbox.TextChanged += new EventHandler(textbox_TextChanged);
            }

            return ypos + textbox.Height + 6;
        }

        void textbox_TextChanged(object sender, EventArgs e)
        {
            short tmp;
            if (short.TryParse(textbox.Text, out tmp))
            {
                val = tmp;
                textbox.Tag = textbox.Text;
            }
            else textbox.Text = textbox.Tag.ToString();
        }
    }

    public class UShortNode : BaseNode
    {
        //Controls
        Label label;
        TextBox textbox;
        Label label2;

        //Values
        ushort val = 0;

        public UShortNode(Map map, StructNode Parent, string Name, string Offset, string Visible)
        {
            this.map = map;
            this.Parent = Parent;
            this.Name = Name;
            this.Offset = int.Parse(Offset);
            this.Visible = bool.Parse(Visible);
            this.Size = 2;
            this.NodeType = NodeType.UShort;
        }

        public override void InitializeControls()
        {


            label = new Label();
            textbox = new TextBox();
            label2 = new Label();


        }

        public override void UpdateVisibility()
        {
            IsVisible = Globals.MetaEditorShownItems[11];
        }

        public override void ReadValue(BinaryReader br, int seekOffset)
        {


            br.BaseStream.Position = (seekOffset == -1 ? map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;
            val = br.ReadUInt16();
            textbox.Tag = val.ToString();
            textbox.Text = val.ToString();


        }

        public override void SaveChanges(BinaryWriter bw, int seekOffset)
        {
            bw.BaseStream.Position = (seekOffset == -1 ?
                map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;

            bw.Write(val);
        }

        public override int DisplayControls(Panel pnl, int xpos, int ypos)
        {


            label.AutoSize = true;
            label.Text = "UShort";
            label.Left = xpos;
            label.Top = ypos;
            pnl.Controls.Add(label);
            xpos += label.Width + 2;

            ypos -= 3;
            textbox.Left = xpos;
            textbox.Top = ypos;
            pnl.Controls.Add(textbox);
            xpos += textbox.Width + 2;

            ypos += 3;
            label2.AutoSize = true;
            label2.Text = this.Name;
            label2.Left = xpos;
            label2.Top = ypos;
            pnl.Controls.Add(label2);

            //cost efficient
            if (Parent == null || Parent.Enabled)
            {
                textbox.TextChanged += new EventHandler(textbox_TextChanged);
            }

            return ypos + textbox.Height + 6;
        }

        void textbox_TextChanged(object sender, EventArgs e)
        {
            ushort tmp;
            if (ushort.TryParse(textbox.Text, out tmp))
            {
                val = tmp;
                textbox.Tag = textbox.Text;
            }
            else textbox.Text = textbox.Tag.ToString();
        }
    }

    public class ByteNode : BaseNode
    {
        //Controls
        Label label;
        TextBox textbox;
        Label label2;

        //Values
        byte val = 0;

        public ByteNode(Map map, StructNode Parent, string Name, string Offset, string Visible)
        {
            this.map = map;
            this.Parent = Parent;
            this.Name = Name;
            this.Offset = int.Parse(Offset);
            this.Visible = bool.Parse(Visible);
            this.Size = 1;
            this.NodeType = NodeType.Byte;
        }

        public override void InitializeControls()
        {


            label = new Label();
            textbox = new TextBox();
            label2 = new Label();


        }

        public override void UpdateVisibility()
        {
            IsVisible = Globals.MetaEditorShownItems[12];
        }

        public override void ReadValue(BinaryReader br, int seekOffset)
        {


            br.BaseStream.Position = (seekOffset == -1 ? map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;
            val = br.ReadByte();
            textbox.Tag = val.ToString();
            textbox.Text = val.ToString();


        }

        public override void SaveChanges(BinaryWriter bw, int seekOffset)
        {
            bw.BaseStream.Position = (seekOffset == -1 ?
                map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;

            bw.Write(val);
        }

        public override int DisplayControls(Panel pnl, int xpos, int ypos)
        {


            label.AutoSize = true;
            label.Text = "Byte";
            label.Left = xpos;
            label.Top = ypos;
            pnl.Controls.Add(label);
            xpos += label.Width + 2;

            ypos -= 3;
            textbox.Left = xpos;
            textbox.Top = ypos;
            pnl.Controls.Add(textbox);
            xpos += textbox.Width + 2;

            ypos += 3;
            label2.AutoSize = true;
            label2.Text = this.Name;
            label2.Left = xpos;
            label2.Top = ypos;
            pnl.Controls.Add(label2);

            //cost efficient
            if (Parent == null || Parent.Enabled)
            {
                textbox.TextChanged += new EventHandler(textbox_TextChanged);
            }

            return ypos + textbox.Height + 6;
        }

        void textbox_TextChanged(object sender, EventArgs e)
        {
            byte tmp;
            if (byte.TryParse(textbox.Text, out tmp))
            {
                val = tmp;
                textbox.Tag = textbox.Text;
            }
            else textbox.Text = textbox.Tag.ToString();
        }
    }

    public class OptionNode : BaseNode
    {
        //Specific
        public int Value;

        public OptionNode(string Name, string Value)
        {
            this.Name = Name;
            this.Value = int.Parse(Value);
            this.NodeType = NodeType.Option;
        }
    }

    public class BitmaskNode : BaseNode
    {
        //Specific
        public List<OptionNode> Nodes = new List<OptionNode>();

        //Controls
        GroupBox groupbox;
        CheckBox[] checkboxes;

        //Values
        object val;

        public BitmaskNode(Map map, StructNode Parent, string Name, string Offset, string Visible, int Size)
        {
            this.map = map;
            this.Parent = Parent;
            this.Name = Name;
            this.Offset = int.Parse(Offset);
            this.Visible = bool.Parse(Visible);
            this.Size = Size;
            this.NodeType = NodeType.Bitmask;
        }

        public override void InitializeControls()
        {
            groupbox = new GroupBox();

            checkboxes = new CheckBox[this.Size];
            for (int x = 0; x < this.Size; x++) checkboxes[x] = new CheckBox();
        }

        public override void UpdateVisibility()
        {
            switch (this.Size)
            {
                case 8:
                    IsVisible = Globals.MetaEditorShownItems[13];
                    break;

                case 16:
                    IsVisible = Globals.MetaEditorShownItems[14];
                    break;

                case 32:
                    IsVisible = Globals.MetaEditorShownItems[15];
                    break;
            }
        }

        public override void ReadValue(BinaryReader br, int seekOffset)
        {
            br.BaseStream.Position = 
                (seekOffset == -1 ? map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;

            switch (this.Size)
            {
                case 8:
                    val = br.ReadByte();
                    break;

                case 16:
                    val = br.ReadUInt16();
                    break;

                case 32:
                    val = br.ReadUInt32();
                    break;
            }

            uint tmp = 1;
            uint tmpVal = Convert.ToUInt32(val);

            for (int x = 0; x < this.Size; x++)
            {
                checkboxes[x].Checked = ((tmp & tmpVal) == tmp);
                tmp <<= 1; //shift
            }
        }

        public override void SaveChanges(BinaryWriter bw, int seekOffset)
        {
            bw.BaseStream.Position = (seekOffset == -1 ?
                map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;

            uint tmp = 1;
            uint tmpVal = 0;

            for (int x = 0; x < this.Size; x++)
            {
                if (checkboxes[x].Checked) tmpVal += tmp;
                tmp <<= 1; //shift
            }
            val = tmpVal;

            switch (this.Size)
            {
                case 8:
                    bw.Write(byte.Parse(val.ToString()));
                    break;

                case 16:
                    bw.Write(ushort.Parse(val.ToString()));
                    break;

                case 32:
                    bw.Write(uint.Parse(val.ToString()));
                    break;
            }
        }

        public override int DisplayControls(Panel pnl, int xpos, int ypos)
        {
            groupbox.Controls.Clear();
            groupbox.Text = this.Name;
            groupbox.Left = xpos;
            groupbox.Top = ypos;
            groupbox.Width = pnl.Width - 24; // (xpos * 2)

            ypos = 20;
            xpos = 6;

            bool secLoop = false;
            string returnStr;
            int longestLen = 0;

            int halfSize =  (Globals.MetaEditorShownItems[4] ?
                this.Size / 2 : (int)Math.Round((double)Nodes.Count / 2, MidpointRounding.AwayFromZero));
          
            int var = 0;
            for (int x = 0; x < this.Size; x++)
            {
                if (ReturnName(x, out returnStr) && !Globals.MetaEditorShownItems[4]) 
                    continue;
                else
                {
                    checkboxes[x].AutoSize = true;
                    checkboxes[x].Text = returnStr;
                    checkboxes[x].Left = xpos;
                    checkboxes[x].Top = ypos;

                    ypos += checkboxes[x].Height;
                    groupbox.Controls.Add(checkboxes[x]);

                    if (!secLoop)
                    {
                        groupbox.Height = ypos;

                        longestLen =
                            checkboxes[x].Width > longestLen ? checkboxes[x].Width : longestLen;

                        var++;
                        if (var == halfSize)
                        {
                            ypos = 20;
                            xpos += longestLen + 6;
                            secLoop = true;
                        }
                    }
                }
            }

            pnl.Controls.Add(groupbox);

            if (Parent != null) groupbox.Anchor = AnchorStyles.Top | AnchorStyles.Left | AnchorStyles.Right;
            else map.entReader.RControls.Add(groupbox);

            return groupbox.Location.Y + groupbox.Height + 12;
        }

        public bool ReturnName(int indx, out string returnStr)
        {
            for (int x = 0; x < Nodes.Count; x++)
            {
                if (Nodes[x].Value == indx)
                {
                    returnStr = Nodes[x].Name;
                    return false;
                }
            }
            returnStr =  "Bit " + indx.ToString();
            return true;
        }
    }

    public class EnumNode : BaseNode
    {
        //Specific
        public List<OptionNode> Nodes = new List<OptionNode>();

        //Controls
        GroupBox groupbox;
        RadioButton[] radiobuttons;

        //Values
        object val;

        public EnumNode(Map map, StructNode Parent, string Name, string Offset, string Visible, int Size)
        {
            this.map = map;
            this.Parent = Parent;
            this.Name = Name;
            this.Offset = int.Parse(Offset);
            this.Visible = bool.Parse(Visible);
            this.Size = Size;
            this.NodeType = NodeType.Enum;
        }

        public override void InitializeControls()
        {
            groupbox = new GroupBox();
            radiobuttons = new RadioButton[this.Nodes.Count];
            for (int x = 0; x < this.Nodes.Count; x++) radiobuttons[x] = new RadioButton();
        }

        public override void UpdateVisibility()
        {
            switch (this.Size)
            {
                case 8:
                    IsVisible = Globals.MetaEditorShownItems[16];
                    break;

                case 16:
                    IsVisible = Globals.MetaEditorShownItems[17];
                    break;

                case 32:
                    IsVisible = Globals.MetaEditorShownItems[18];
                    break;
            }
        }

        public override void ReadValue(BinaryReader br, int seekOffset)
        {
            br.BaseStream.Position = (seekOffset == -1 ? map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;

            switch (this.Size)
            {
                case 8:
                    val = br.ReadByte();
                    if ((byte)val < radiobuttons.Length &&
                        (byte)val >= 0) radiobuttons[(byte)val].Checked = true;
                    break;

                case 16:
                    val = br.ReadInt16();
                    if ((short)val < radiobuttons.Length &&
                        (short)val >= 0) radiobuttons[(short)val].Checked = true;
                    break;

                case 32:
                    val = br.ReadInt32();
                    if ((int)val < radiobuttons.Length &&
                        (int)val >= 0) radiobuttons[(int)val].Checked = true;
                    break;
            }
        }

        public override void SaveChanges(BinaryWriter bw, int seekOffset)
        {
            bw.BaseStream.Position = (seekOffset == -1 ?
                map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;

            switch (this.Size)
            {
                case 8:
                    bw.Write(byte.Parse(val.ToString()));
                    break;

                case 16:
                    bw.Write(short.Parse(val.ToString()));
                    break;

                case 32:
                    bw.Write(int.Parse(val.ToString()));
                    break;
            }
        }

        public override int DisplayControls(Panel pnl, int xpos, int ypos)
        {
            groupbox.Controls.Clear();
            groupbox.Text = this.Name;
            groupbox.Left = xpos;
            groupbox.Top = ypos;
            groupbox.Width = pnl.Width - 24; // (xpos * 2)

            ypos = 20;
            xpos = 6;

            bool secLoop = false;
            int longestLen = 0;

            int halfSize = (int)Math.Round((double)Nodes.Count / 2, MidpointRounding.AwayFromZero) - 1;

            for (int x = 0; x < Nodes.Count; x++)
            {
                radiobuttons[x].AutoSize = true;
                radiobuttons[x].Text = Nodes[x].Name;
                radiobuttons[x].Left = xpos;
                radiobuttons[x].Top = ypos;

                ypos += radiobuttons[x].Height;
                groupbox.Controls.Add(radiobuttons[x]);

                if (!secLoop)
                {
                    longestLen =
                        radiobuttons[x].Width > longestLen ? radiobuttons[x].Width : longestLen;

                    if (x == halfSize)
                    {
                        groupbox.Height = ypos;

                        ypos = 20;
                        xpos += longestLen + 6;
                        secLoop = true;
                    }
                }
            }

            pnl.Controls.Add(groupbox);

            //cost efficient
            if (Parent == null || Parent.Enabled)
            {
                for (int x = 0; x < radiobuttons.Length; x++)
                    radiobuttons[x].CheckedChanged += new EventHandler(radiobutton_CheckedChanged);
            }

            if (Parent != null) groupbox.Anchor = AnchorStyles.Top | AnchorStyles.Left | AnchorStyles.Right;
            else map.entReader.RControls.Add(groupbox);

            return groupbox.Location.Y + groupbox.Height + 12;
        }

        void radiobutton_CheckedChanged(object sender, EventArgs e)
        {
            if ((sender as RadioButton).Checked)
                val = groupbox.Controls.IndexOf(sender as RadioButton);
        }
    }

    public class StringNode : BaseNode
    {
        //Controls
        Label label;
        TextBox textbox;
        Label label2;

        //Values
        public string sString;

        public StringNode(Map map, StructNode Parent, string Name, string Offset, string Visible, int Size)
        {
            this.map = map;
            this.Parent = Parent;
            this.Name = Name;
            this.Offset = int.Parse(Offset);
            this.Visible = bool.Parse(Visible);
            this.Size = Size;
            this.NodeType = NodeType.String;
        }

        public override void InitializeControls()
        {


            label = new Label();
            textbox = new TextBox();
            label2 = new Label();


        }

        public override void UpdateVisibility()
        {
            switch (this.Size)
            {
                case 32:
                    IsVisible = Globals.MetaEditorShownItems[19];
                    break;

                case 256:
                    IsVisible = Globals.MetaEditorShownItems[20];
                    break;
            }
        }

        public override void ReadValue(BinaryReader br, int seekOffset)
        {


            br.BaseStream.Position = (seekOffset == -1 ? map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;

            char chr;
            sString = "";

            while ((chr = br.ReadChar()) != '\0')
                sString += chr;

            textbox.Text = sString;
            textbox.MaxLength = this.Size; // mayb?


        }

        public override void SaveChanges(BinaryWriter bw, int seekOffset)
        {
            bw.BaseStream.Position = (seekOffset == -1 ?
                map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;

            bw.Write(sString.ToCharArray());
            byte[] nul = new byte[this.Size - sString.Length];
            bw.Write(nul);
        }

        public override int DisplayControls(Panel pnl, int xpos, int ypos)
        {


            label.AutoSize = true;
            label.Text = "String" + this.Size.ToString();
            label.Left = xpos;
            label.Top = ypos;
            pnl.Controls.Add(label);
            xpos += label.Width + 2;

            ypos -= 3;
            textbox.Left = xpos;
            textbox.Top = ypos;
            pnl.Controls.Add(textbox);
            xpos += textbox.Width + 2;

            ypos += 3;
            label2.AutoSize = true;
            label2.Text = this.Name;
            label2.Left = xpos;
            label2.Top = ypos;
            pnl.Controls.Add(label2);

            //cost efficient
            if (Parent == null || Parent.Enabled)
            {
                textbox.TextChanged += new EventHandler(textbox_TextChanged);
            }

            return ypos + textbox.Height + 6;
        }

        void textbox_TextChanged(object sender, EventArgs e)
        {
            sString = textbox.Text;
        }
    }

    public class UnicodeNode : BaseNode
    {
        //Controls
        Label label;
        TextBox textbox;
        Label label2;

        //Values
        public string sUnicode;

        public UnicodeNode(Map map, StructNode Parent, string Name, string Offset, string Visible, int Size)
        {
            this.map = map;
            this.Parent = Parent;
            this.Name = Name;
            this.Offset = int.Parse(Offset);
            this.Visible = bool.Parse(Visible);
            this.Size = Size;
            this.NodeType = NodeType.Unicode;
        }

        public override void InitializeControls()
        {
            label = new Label();
            textbox = new TextBox();
            label2 = new Label();
        }

        public override void UpdateVisibility()
        {
            switch (this.Size)
            {
                case 64:
                    IsVisible = Globals.MetaEditorShownItems[21];
                    break;

                case 256:
                    IsVisible = Globals.MetaEditorShownItems[22];
                    break;
            }
        }

        public override void ReadValue(BinaryReader br, int seekOffset)
        {
            br.BaseStream.Position = (seekOffset == -1 ? map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;

            Encoding dec = Encoding.Unicode;
            sUnicode = dec.GetString(br.ReadBytes(this.Size));

            //remove nulling at end
            sUnicode = sUnicode.Remove(sUnicode.IndexOf('\0'));

            textbox.Text = sUnicode;
            textbox.MaxLength = this.Size; // mayb?
        }

        public override void SaveChanges(BinaryWriter bw, int seekOffset)
        {
            bw.BaseStream.Position = (seekOffset == -1 ?
                map.tags.tagsList[map.tags.selectedIndex].MetaOffset : seekOffset) + Offset;

            Encoding dec = Encoding.Unicode;
            byte[] bUnicode = dec.GetBytes(sUnicode);
            bw.Write(bUnicode);

            byte[] nul = new byte[this.Size - bUnicode.Length];
            bw.Write(nul);
        }

        public override int DisplayControls(Panel pnl, int xpos, int ypos)
        {


            label.AutoSize = true;
            label.Text = "Unicode" + this.Size.ToString();
            label.Left = xpos;
            label.Top = ypos;
            pnl.Controls.Add(label);
            xpos += label.Width + 2;

            ypos -= 3;
            textbox.Left = xpos;
            textbox.Top = ypos;
            pnl.Controls.Add(textbox);
            xpos += textbox.Width + 2;

            ypos += 3;
            label2.AutoSize = true;
            label2.Text = this.Name;
            label2.Left = xpos;
            label2.Top = ypos;
            pnl.Controls.Add(label2);

            //cost efficient
            if (Parent == null || Parent.Enabled)
            {
                textbox.TextChanged += new EventHandler(textbox_TextChanged);
            }

            return ypos + textbox.Height + 6;
        }

        void textbox_TextChanged(object sender, EventArgs e)
        {
            sUnicode = textbox.Text;
        }
    }
}