using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace Dark
{
    public struct PaletteColor
    {
        public int a, r, g, b;
    }

    public class LightMap
    {
        public int Ident;
        public int TagIndex;
        public int Offset;
        public int Size;

        public int PalettesOffset;
        public List<PaletteColor[]> Palettes = new List<PaletteColor[]>();

        public int[]
        VisualChunk_LightMap_Index,
        VisualChunk_Bitmap_Index,
        SceneryChunk_LightMap_Index,
        SceneryChunk_Bitmap_Index;

        public void WritePalettes(string path)
        {
            BinaryWriter bw = new BinaryWriter(new FileStream(path,
                    FileMode.Open, FileAccess.ReadWrite, FileShare.ReadWrite));

            bw.BaseStream.Position = this.PalettesOffset;
            foreach (PaletteColor[] pal in this.Palettes)
            {
                for (int x = 0; x < 256; x++)
                {
                    bw.Write((byte)pal[x].r);
                    bw.Write((byte)pal[x].g);
                    bw.Write((byte)pal[x].b);
                    bw.Write((byte)pal[x].a);
                }
            }
            bw.Close();
        }
    }

    public class SbspInfo
    {
        public int Ident;
        public int PointerOffset;
        public int Offset;
        public int Size;
        public int TagIndex;
        public int Magic;
        public LightMap lightmap = new LightMap();
    }

    public class Sbsp
    {
        public Map map;
        public SbspInfo[] sbspInfoList;

        public Sbsp(Map map, BinaryReader br)
        {
            this.map = map;

            //get the scnr index
            int scnrIndx = map.tags.FindTagIndexByID(map.index.scnrID);
            if (scnrIndx == -1) return;

            //get the scnr offset
            int scnrOffset = map.tags.tagsList[scnrIndx].MetaOffset;

            br.BaseStream.Position = scnrOffset + 528;

            //chunk count and translated reflexive
            int bspCount = br.ReadInt32(),
                bspTransReflx = br.ReadInt32() - map.index.secondaryMagic;

            sbspInfoList = new SbspInfo[bspCount];

            //count through the bsp's in the map
            for (int x = 0; x < bspCount; x++)
            {
                //for storing values to come
                sbspInfoList[x] = new SbspInfo();
                sbspInfoList[x].PointerOffset = bspTransReflx + (x * 68) - scnrOffset;

                //each chunk is 68 bytes long
                br.BaseStream.Position = bspTransReflx + (x * 68);
                sbspInfoList[x].Offset = br.ReadInt32() & (int)0x3FFFFFFF;

                //size and magic
                sbspInfoList[x].Size = br.ReadInt32();
                sbspInfoList[x].Magic = br.ReadInt32() - sbspInfoList[x].Offset;

                br.BaseStream.Position += 8;
                sbspInfoList[x].Ident = br.ReadInt32();
                sbspInfoList[x].TagIndex = map.tags.FindTagIndexByID(sbspInfoList[x].Ident);

                //fix the current bsp's meta offset and size
                map.tags.tagsList[sbspInfoList[x].TagIndex].MetaOffset = sbspInfoList[x].Offset;
                map.tags.tagsList[sbspInfoList[x].TagIndex].MetaSize = sbspInfoList[x].Size;

                br.BaseStream.Position += 4;
                sbspInfoList[x].lightmap.Ident = br.ReadInt32();
                sbspInfoList[x].lightmap.TagIndex = map.tags.FindTagIndexByID(sbspInfoList[x].lightmap.Ident);

                //this would mean a broken ident
                if (sbspInfoList[x].lightmap.TagIndex == -1)
                    continue; //on to the next loop

                br.BaseStream.Position = sbspInfoList[x].Offset + 8;
                sbspInfoList[x].lightmap.Offset = br.ReadInt32();

                //may mean bsp is not linked to a lightmap
                if (sbspInfoList[x].lightmap.Offset == 0)
                    continue;

                sbspInfoList[x].lightmap.Offset -= sbspInfoList[x].Magic;
                sbspInfoList[x].lightmap.Size = (sbspInfoList[x].Offset + sbspInfoList[x].Size) - sbspInfoList[x].lightmap.Offset;

                //fix the current bsp's ltmp meta offset and size
                map.tags.tagsList[sbspInfoList[x].lightmap.TagIndex].MetaOffset = sbspInfoList[x].lightmap.Offset;
                map.tags.tagsList[sbspInfoList[x].lightmap.TagIndex].MetaSize = sbspInfoList[x].lightmap.Size;

                //lightmap bitmap
                br.BaseStream.Position = sbspInfoList[x].lightmap.Offset + 128;

                int tempc = br.ReadInt32();
                int temprx = br.ReadInt32() - sbspInfoList[x].Magic;

                //br.BaseStream.Position = temprx + 28;
                //sbspInfoList[x].lightmap.TagIndex = map.tags.FindTagIndexByID(br.ReadInt32());

                //light map palettes
                br.BaseStream.Position = temprx + 8;

                int tempc2 = br.ReadInt32();
                int tempr2 = br.ReadInt32() - sbspInfoList[x].Magic;

                sbspInfoList[x].lightmap.PalettesOffset = tempr2;

                for (int y = 0; y < tempc2; y++)
                {
                    br.BaseStream.Position = tempr2 + (y * 1024);

                    PaletteColor[] pc = new PaletteColor[256];

                    for (int z = 0; z < 256; z++)
                    {
                        pc[z] = new PaletteColor();
                        pc[z].r = br.ReadByte();
                        pc[z].g = br.ReadByte();
                        pc[z].b = br.ReadByte();
                        pc[z].a = br.ReadByte();
                    }
                    sbspInfoList[x].lightmap.Palettes.Add(pc);
                }

                br.BaseStream.Position = temprx + 40;

                tempc2 = br.ReadInt32();
                tempr2 = br.ReadInt32() - sbspInfoList[x].Magic;

                if (tempc2 != 0)
                {
                    sbspInfoList[x].lightmap.VisualChunk_Bitmap_Index = new int[tempc2];
                    sbspInfoList[x].lightmap.VisualChunk_LightMap_Index = new int[tempc2];

                    br.BaseStream.Position = tempr2;

                    for (int y = 0; y < tempc2; y++)
                    {
                        sbspInfoList[x].lightmap.VisualChunk_Bitmap_Index[y] = br.ReadInt16();
                        sbspInfoList[x].lightmap.VisualChunk_LightMap_Index[y] = br.ReadInt16();
                    }
                }

                br.BaseStream.Position = temprx + 72;

                tempc2 = br.ReadInt32();
                tempr2 = br.ReadInt32() - sbspInfoList[x].Magic;

                sbspInfoList[x].lightmap.SceneryChunk_Bitmap_Index = new int[tempc2];
                sbspInfoList[x].lightmap.SceneryChunk_LightMap_Index = new int[tempc2];

                if (tempc2 > 0)
                {
                    br.BaseStream.Position = tempr2;
                    for (int y = 0; y < tempc2; y++)
                    {
                        sbspInfoList[x].lightmap.SceneryChunk_Bitmap_Index[y] = br.ReadInt16();
                        sbspInfoList[x].lightmap.SceneryChunk_LightMap_Index[y] = br.ReadInt16();
                    }
                }
            }
        }

        public int FindIndexByTagID(int ID)
        {
            for (int x = 0; x < sbspInfoList.Length; x++)
            {
                if (sbspInfoList[x].Ident == ID)
                    return x;
            }
            return -1;
        }

        public int FindIndexByLightmapTagID(int ID)
        {
            for (int x = 0; x < sbspInfoList.Length; x++)
            {
                if (sbspInfoList[x].lightmap.Ident == ID)
                    return x;
            }
            return -1;
        }
    }
}
