using System;
using System.Collections.Generic;
using System.Text;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.IO;
using System.Windows.Forms;

namespace Visual_Scenery_Mover
{
    public struct HaloBSPVertex
    {
        public Vector3
        Position, Normal;

        public int diffuse;
        // public int specular;

        public float
        Tu0, Tv0,
        Tu1, Tv1,
        Tu2, Tv2,
        Tu3, Tv3;

        public static readonly VertexFormats FVF = VertexFormats.PositionNormal |
            VertexFormats.Diffuse | VertexFormats.Texture4;
    }

    public class BSPModel : IDisposable
    {
        public int BSPRawDataMetaChunksOffset;
        public BSPRawDataMetaChunk[] BSPRawDataMetaChunks;
        public int BSPPermutationRawDataMetaChunksOffset;
        public BSPPermutationRawDataMetaChunk[] BSPPermutationRawDataMetaChunks;
        public PermutationPlacement[] PermutationInfo;
        public BSPShaderContainer Shaders;
        public BSPDisplayedInfo Display;
        public Sky sky;
        public ParsedModel SkyBox;
        public SpawnInfo Spawns;
        public string Name;
        public UnknownChunk[] UnknownChunks;
        public ClusterPVS[] ClusterInfo;
        public ParsedBitmap LightMapParsedBitmap;
        public Texture[] LightMapTexture;
        public System.Drawing.Bitmap[] LightMapBitmap;
        public Texture[] SceneryLightMapTexture;
        public System.Drawing.Bitmap[] SceneryLightMapBitmap;
        public int BspNumber;
        public bool renderLightmap, renderSkybox;
       // public Mesh[] BoundingBoxScenery;
        public Mesh[] SceneryMesh;

        public BSPModel(int mapnum)
        {
            renderLightmap = Pref.renderLightmap;
            renderSkybox = Pref.renderSkybox;

            Meta meta = Map.mp[mapnum].meta;

            Name = meta.name.Substring(meta.name.LastIndexOf('\\') + 1);
            BspNumber = Map.mp[mapnum].bsp.FindBSPNumberByBSPID(meta.ident);

            LoadModelStructure(ref meta);
            LoadPermutations(ref meta);
            LoadUnknowns(ref meta);
            Shaders = new BSPShaderContainer(this, mapnum);
            Display = new BSPDisplayedInfo();
            LoadLightmaps(mapnum);
            sky = new Sky();
            LoadSky(mapnum);
            Spawns = new SpawnInfo(Map.mp[mapnum].mapLoc, Map.mp[mapnum].tags, Map.mp[mapnum].index.secondaryMagic);
            ClusterInfo = ClustersPVS.GetClustersPVSInfo(ref meta);
        }

        public void Dispose()
        {
            this.BSPPermutationRawDataMetaChunks = null;
            this.PermutationInfo = null;
            this.BSPRawDataMetaChunks = null;
            this.Display = null;

            this.Shaders = null;
            this.SkyBox = null;
            this.Spawns = null;
            GC.SuppressFinalize(this);
        }

        public static void AddUncompressedVertice(ref BinaryReader br, ref BSPRawDataMetaChunk chunk)
        {
            chunk.Vertices.Add(new Vector3(br.ReadSingle(), br.ReadSingle(), br.ReadSingle()));
            chunk.Normals.Add(new Vector3(br.ReadSingle(), br.ReadSingle(), br.ReadSingle()));
            chunk.Binormals.Add(new Vector3(br.ReadSingle(), br.ReadSingle(), br.ReadSingle()));
            chunk.Tangents.Add(new Vector3(br.ReadSingle(), br.ReadSingle(), br.ReadSingle()));
            chunk.UVs.Add(new Vector2(br.ReadSingle(), br.ReadSingle()));
        }

        public static void AddCompressedVertice(ref BinaryReader BR, ref BSPRawDataMetaChunk chunk)
        {
            chunk.Vertices.Add(new Vector3(BR.ReadSingle(), BR.ReadSingle(), BR.ReadSingle()));
            chunk.Normals.Add(ParsedModel.DecompressNormal(BR.ReadInt32()));
            chunk.Binormals.Add(ParsedModel.DecompressNormal(BR.ReadInt32()));
            chunk.Tangents.Add(ParsedModel.DecompressNormal(BR.ReadInt32()));
            //---------------basemap uvs
            chunk.UVs.Add(new Vector2(BR.ReadSingle(), BR.ReadSingle()));

            //---------------lightmap uvs
            // testx = BR.ReadInt16();
            // float ux = chunk.DecompressVertice(Convert.ToSingle(testx), -1, 1);// %1 ;
            // testx = BR.ReadInt16();
            // float vx =chunk.DecompressVertice(Convert.ToSingle(testx), -1, 1);// % 1;
            // Vector2 uv2x = new Vector2(ux, vx);
            // chunk.LightMapUVs.Add(uv2);
        }

        public class Face
        {
            public int
            vertex0Index,
            vertex1Index,
            vertex2Index;
        }

        public void LoadSky(int mapnum)
        {
            //open the map
            BinaryReader br = new BinaryReader(new FileStream(Map.mp[mapnum].mapLoc,
                    FileMode.Open, FileAccess.Read, FileShare.ReadWrite));

            //scnr offset + 8
            br.BaseStream.Position = Map.mp[mapnum].tags.tagInfo[3].MetaOffset + 8;

            int tempc = br.ReadInt32();
            if (tempc == 0) return;

            int tempr = br.ReadInt32() - Map.mp[mapnum].index.secondaryMagic;
            br.BaseStream.Position = tempr + 4;

            int tempident = Meta.FindMetaByID(br.ReadInt32(), Map.mp[mapnum].tags.tagInfo);
            if (tempident != -1)
            {
                sky = new Sky(tempident, Map.mp[mapnum].mapLoc, Map.mp[mapnum].tags.tagInfo, Map.mp[mapnum].index.secondaryMagic);

                br.BaseStream.Position = Map.mp[mapnum].tags.tagInfo[tempident].MetaOffset + 4;
                tempident = Meta.FindMetaByID(br.ReadInt32(), Map.mp[mapnum].tags.tagInfo);
                if (tempident != -1)
                {
                    Meta tempmeta = new Meta(tempident);
                    tempmeta.ReadMetaFromMap(false, true, false, mapnum);
                    SkyBox = new ParsedModel(ref tempmeta, mapnum);

                }
            }
            br.Close();
        }

        public class BSPRawDataMetaChunk : ParsedModel.RawDataMetaChunk
        {
            public int rawdatainfooffset, shadertagnumber;
            public List<Vector2> LightMapUVs = new List<Vector2>();

            public BSPRawDataMetaChunk()
                : base()
            { }

            public BSPRawDataMetaChunk(int offset, int chunknumber, ref Meta meta, Face[] faces)
                : base()
            {
                BinaryReader br = new BinaryReader(meta.MS);
                br.BaseStream.Position = offset;

                this.VerticeCount = br.ReadUInt16();
                this.FaceCount = br.ReadUInt16();

                br.BaseStream.Position = offset + 48;

                this.HeaderSize = br.ReadInt32() + 8;
                br.ReadInt32();
                int tempc = br.ReadInt32();
                int tempr = br.ReadInt32() - meta.magic - meta.offset;

                this.RawDataChunkInfo = new RawDataOffsetChunk[tempc];

                rawdatainfooffset = tempr;
                for (int x = 0; x < tempc; x++)
                {
                    this.RawDataChunkInfo[x] = new RawDataOffsetChunk();
                    br.BaseStream.Position = tempr + (x * 16) + 6;

                    this.RawDataChunkInfo[x].ChunkSize = br.ReadUInt16();
                    this.RawDataChunkInfo[x].Size = br.ReadInt32();
                    if (this.RawDataChunkInfo[x].ChunkSize == 0)
                        this.RawDataChunkInfo[x].ChunkSize = this.RawDataChunkInfo[x].Size;

                    this.RawDataChunkInfo[x].Offset = br.ReadInt32();
                    this.RawDataChunkInfo[x].ChunkCount = this.RawDataChunkInfo[x].Size / this.RawDataChunkInfo[x].ChunkSize;
                }

                if (this.RawDataChunkInfo.Length == 0) return;

                br = new BinaryReader(((RawDataChunk)meta.raw.rawChunks[chunknumber]).MS);

                this.SubMeshInfo = new ModelSubMeshInfo[this.RawDataChunkInfo[0].ChunkCount];
                for (int x = 0; x < this.RawDataChunkInfo[0].ChunkCount; x++)
                {
                    this.SubMeshInfo[x] = new ModelSubMeshInfo();
                    br.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[0].Offset + (x * 72) + 4;
                    this.SubMeshInfo[x].ShaderNumber = br.ReadUInt16();
                    this.SubMeshInfo[x].IndiceStart = br.ReadUInt16();
                    this.SubMeshInfo[x].IndiceCount = br.ReadUInt16();
                }

                br.BaseStream.Position = 40;
                this.IndiceCount = br.ReadUInt16();

                int indicechunk = 0, verticechunk = 0, uvchunk = 0;

                for (int x = 0; x < RawDataChunkInfo.Length; x++)
                {
                    if (RawDataChunkInfo[x].ChunkSize == 2)
                    {
                        indicechunk = x;
                        break;
                    }
                }

                int normalchunk = 0;
                for (int x = indicechunk; x < RawDataChunkInfo.Length; x++)
                {
                    if (RawDataChunkInfo[x].ChunkCount == 1)
                    {
                        verticechunk = x;
                        uvchunk = x + 1;
                        normalchunk = x + 2;
                        break;
                    }
                }
                this.Indices = new List<short>();
                br.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[indicechunk].Offset;
                for (int x = 0; x < this.IndiceCount; x++)
                    this.Indices.Add((short)br.ReadUInt16());

                this.RawDataChunkInfo[verticechunk].ChunkSize = this.RawDataChunkInfo[verticechunk].Size / VerticeCount;
                for (int x = 0; x < this.VerticeCount; x++)
                {
                    br.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[verticechunk].Offset + (this.RawDataChunkInfo[verticechunk].ChunkSize * x);
                    Vertices.Add(new Vector3(br.ReadSingle(), br.ReadSingle(), br.ReadSingle()));
                }

                this.RawDataChunkInfo[uvchunk].ChunkSize = 8;
                for (int x = 0; x < this.VerticeCount; x++)
                {
                    br.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[uvchunk].Offset + (this.RawDataChunkInfo[uvchunk].ChunkSize * x);
                    this.UVs.Add(new Vector2(br.ReadSingle(), br.ReadSingle()));
                }

                this.RawDataChunkInfo[normalchunk].ChunkSize = 12;
                for (int x = 0; x < this.VerticeCount; x++)
                {
                    br.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[normalchunk].Offset + (this.RawDataChunkInfo[normalchunk].ChunkSize * x);

                    this.Normals.Add(ParsedModel.DecompressNormal(br.ReadInt32()));
                    this.Binormals.Add(ParsedModel.DecompressNormal(br.ReadInt32()));
                    this.Tangents.Add(ParsedModel.DecompressNormal(br.ReadInt32()));
                }

                int lightmapuvchunk = -1;
                for (int x = normalchunk + 1; x < RawDataChunkInfo.Length; x++)
                {
                    if (RawDataChunkInfo[x].ChunkSize == 3)
                    {
                        lightmapuvchunk = x;
                        break;
                    }
                }

                if (lightmapuvchunk == -1) return;
                RawDataChunkInfo[lightmapuvchunk].ChunkSize = 4;
                for (int x = 0; x < this.VerticeCount; x++)
                {
                    br.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[lightmapuvchunk].Offset + (this.RawDataChunkInfo[lightmapuvchunk].ChunkSize * x);

                    this.LightMapUVs.Add(new Vector2(DecompressVertice(Convert.ToSingle(br.ReadInt16()), -1, 1),
                        DecompressVertice(Convert.ToSingle(br.ReadInt16()), -1, 1)));
                }
            }
        }

        public void LoadModelStructure(ref Meta meta)
        {
            BinaryReader br = new BinaryReader(meta.MS);
            br.BaseStream.Position = 172;

            int tempc = br.ReadInt32(), tempr = br.ReadInt32() - meta.magic - meta.offset;
            BSPRawDataMetaChunks = new BSPRawDataMetaChunk[tempc];
            BSPRawDataMetaChunksOffset = tempr;

            for (int x = 0; x < tempc; x++)
                BSPRawDataMetaChunks[x] = new BSPRawDataMetaChunk(tempr + (x * 176), x, ref meta, null);
        }

        public class UnknownChunk
        {
            public ParsedModel.RawDataMetaChunk.RawDataOffsetChunk[] RawDataChunkInfo;
            public int
            VerticeCount,
            HeaderSize,
            IndiceCount,
            facecount;

            public short[] Indices;
            public List<Vector3> Vertices = new List<Vector3>();

            public UnknownChunk(int offset, int chunknumber, ref Meta meta)
            {
                BinaryReader br = new BinaryReader(meta.MS);
                br.BaseStream.Position = offset;

                br.BaseStream.Position = offset + 8;
                this.HeaderSize = br.ReadInt32() + 8;

                // LoadRawDataChunkInfo
                br.BaseStream.Position = offset + 16;
                int tempc = br.ReadInt32(), tempr = br.ReadInt32() - meta.magic - meta.offset;

                this.RawDataChunkInfo = new ParsedModel.RawDataMetaChunk.RawDataOffsetChunk[tempc];
                for (int x = 0; x < tempc; x++)
                {
                    this.RawDataChunkInfo[x] = new ParsedModel.RawDataMetaChunk.RawDataOffsetChunk();
                    br.BaseStream.Position = tempr + (x * 16) + 6;

                    this.RawDataChunkInfo[x].ChunkSize = br.ReadUInt16();
                    this.RawDataChunkInfo[x].Size = br.ReadInt32();

                    if (this.RawDataChunkInfo[x].ChunkSize == 0)
                        this.RawDataChunkInfo[x].ChunkSize = this.RawDataChunkInfo[x].Size;

                    this.RawDataChunkInfo[x].Offset = br.ReadInt32();
                    this.RawDataChunkInfo[x].ChunkCount = this.RawDataChunkInfo[x].Size / this.RawDataChunkInfo[x].ChunkSize;
                }
                if (this.RawDataChunkInfo.Length == 0) return;

                // RawData
                int firstbsp2chunk = 0;

                // FindFirstRawData
                for (int x = 0; x < meta.raw.rawChunks.Count; x++)
                {
                    if (((RawDataChunk)meta.raw.rawChunks[x]).rawDataType == RawDataType.bsp3)
                    {
                        firstbsp2chunk = x;
                        break;
                    }
                }

                br = new BinaryReader(((RawDataChunk)meta.raw.rawChunks[firstbsp2chunk + chunknumber]).MS);

                // DetermineChunkNumbers
                int indicechunk = 0, verticechunk = 0;

                for (int x = 0; x < RawDataChunkInfo.Length; x++)
                {
                    if (RawDataChunkInfo[x].ChunkSize == 32)
                    {
                        verticechunk = x;
                        indicechunk = x + 1;
                        break;
                    }
                }

                // Faces
                this.IndiceCount = this.RawDataChunkInfo[indicechunk].ChunkCount;
                this.Indices = new short[this.IndiceCount];

                br.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[indicechunk].Offset;
                for (int x = 0; x < this.IndiceCount; x++)
                    this.Indices[x] = (short)br.ReadUInt16();

                // Vertices
                this.VerticeCount = this.RawDataChunkInfo[verticechunk].Size / 32;

                for (int x = 0; x < this.VerticeCount; x++)
                {
                    br.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[verticechunk].Offset + (this.RawDataChunkInfo[verticechunk].ChunkSize * x);
                    Vertices.Add(new Vector3(br.ReadSingle(), br.ReadSingle(), br.ReadSingle()));
                }
            }
        }

        public void LoadUnknowns(ref Meta meta)
        {
            BinaryReader br = new BinaryReader(meta.MS);
            br.BaseStream.Position = 580;

            int tempc = br.ReadInt32(), tempr = br.ReadInt32() - meta.magic - meta.offset;
            br.BaseStream.Position = tempr + 16;

            tempc = br.ReadInt32();
            tempr = br.ReadInt32() - meta.magic - meta.offset;

            UnknownChunks = new UnknownChunk[tempc];
            for (int x = 0; x < tempc; x++)
                UnknownChunks[x] = new UnknownChunk(tempr + (x * 44), x, ref meta);
        }

        public class BSPPermutationRawDataMetaChunk : ParsedModel.RawDataMetaChunk
        {
            public BSPPermutationBoundingBoxContainer BoundingBox;
            public Mesh Box;

            public List<Vector2> LightMapUVs = new List<Vector2>();

            public BSPPermutationRawDataMetaChunk(int offset, int chunknumber, ref Meta meta)
            {
                BinaryReader br = new BinaryReader(meta.MS);
                br.BaseStream.Position = offset;
                this.VerticeCount = br.ReadUInt16();
                this.FaceCount = br.ReadUInt16();

                // LoadBoundingBox
                br.BaseStream.Position = offset + 24;
                int tempc = br.ReadInt32();
                int tempr = br.ReadInt32() - meta.magic - meta.offset;
                br.BaseStream.Position = tempr;
                //BoundingBox = new BSPPermutationBoundingBoxContainer(ref br, ref Box);


                br.BaseStream.Position = offset + 48;
                this.HeaderSize = br.ReadInt32() + 8;

                // LoadRawDataChunkInfo
                br.BaseStream.Position = offset + 56;
                tempc = br.ReadInt32();
                tempr = br.ReadInt32() - meta.magic - meta.offset;
                this.RawDataChunkInfo = new RawDataOffsetChunk[tempc];
                for (int x = 0; x < tempc; x++)
                {
                    this.RawDataChunkInfo[x] = new RawDataOffsetChunk();
                    br.BaseStream.Position = tempr + (x * 16) + 6;

                    this.RawDataChunkInfo[x].ChunkSize = br.ReadUInt16();
                    this.RawDataChunkInfo[x].Size = br.ReadInt32();

                    if (this.RawDataChunkInfo[x].ChunkSize == 0)
                        this.RawDataChunkInfo[x].ChunkSize = this.RawDataChunkInfo[x].Size;

                    this.RawDataChunkInfo[x].Offset = br.ReadInt32();
                    this.RawDataChunkInfo[x].ChunkCount = this.RawDataChunkInfo[x].Size / this.RawDataChunkInfo[x].ChunkSize;
                }
                if (this.RawDataChunkInfo.Length == 0) return;

                // RawData
                int firstbsp2chunk = 0;

                // FindFirstPermutationRawData
                for (int x = 0; x < meta.raw.rawChunks.Count; x++)
                {
                    if (((RawDataChunk)meta.raw.rawChunks[x]).rawDataType == RawDataType.bsp2)
                    {
                        firstbsp2chunk = x;
                        break;
                    }
                }

                // LoadSubMeshInfo
                br = new BinaryReader(((RawDataChunk)meta.raw.rawChunks[firstbsp2chunk + chunknumber]).MS);

                this.SubMeshInfo = new ModelSubMeshInfo[this.RawDataChunkInfo[0].ChunkCount];
                for (int x = 0; x < this.RawDataChunkInfo[0].ChunkCount; x++)
                {
                    this.SubMeshInfo[x] = new ModelSubMeshInfo();
                    br.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[0].Offset + (x * 72) + 4;
                    this.SubMeshInfo[x].ShaderNumber = br.ReadUInt16();
                    this.SubMeshInfo[x].IndiceStart = br.ReadUInt16();
                    this.SubMeshInfo[x].IndiceCount = br.ReadUInt16();
                }

                // DetermineChunkNumers
                int
                indicechunk = 0, verticechunk = 0,
                uvchunk = 0, normalchunk = 0;
                for (int x = 0; x < RawDataChunkInfo.Length; x++)
                {
                    if (RawDataChunkInfo[x].ChunkSize == 2)
                    {
                        indicechunk = x;
                        break;
                    }
                }

                for (int x = indicechunk; x < RawDataChunkInfo.Length; x++)
                {
                    if (RawDataChunkInfo[x].ChunkCount == 1)
                    {
                        verticechunk = x;
                        uvchunk = x + 1;
                        normalchunk = x + 2;
                        break;
                    }
                }

                // Faces
                br.BaseStream.Position = 40;
                this.IndiceCount = br.ReadUInt16();
                this.Indices = new List<short>();

                br.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[indicechunk].Offset;
                for (int x = 0; x < this.IndiceCount; x++)
                    this.Indices.Add((short)br.ReadUInt16());

                // Vertices
                this.RawDataChunkInfo[verticechunk].ChunkSize = this.RawDataChunkInfo[verticechunk].Size / this.VerticeCount;

                for (int x = 0; x < this.VerticeCount; x++)
                {
                    br.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[verticechunk].Offset + (this.RawDataChunkInfo[verticechunk].ChunkSize * x);
                    Vertices.Add(new Vector3(br.ReadSingle(), br.ReadSingle(), br.ReadSingle()));
                }

                // UV's
                this.RawDataChunkInfo[uvchunk].ChunkSize = 8;
                for (int x = 0; x < this.VerticeCount; x++)
                {
                    br.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[uvchunk].Offset + (this.RawDataChunkInfo[uvchunk].ChunkSize * x);

                    //if (this.UVs[x].X > 1)  this.UVs[x].X = this.UVs[x].X - 1; 
                    //if (this.UVs[x].X < 0)  this.UVs[x].X = this.UVs[x].X + 1; 
                    //if (this.UVs[x].Y > 1)  this.UVs[x].Y = this.UVs[x].Y - 1; 
                    //if (this.UVs[x].Y < 0)  this.UVs[x].Y = this.UVs[x].Y + 1; 
                    this.UVs.Add(new Vector2(br.ReadSingle(), br.ReadSingle()));
                }

                // Normals
                this.RawDataChunkInfo[normalchunk].ChunkSize = 12;
                for (int x = 0; x < this.VerticeCount; x++)
                {
                    br.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[normalchunk].Offset + (this.RawDataChunkInfo[normalchunk].ChunkSize * x);

                    this.Normals.Add(ParsedModel.DecompressNormal(br.ReadInt32()));
                    this.Binormals.Add(ParsedModel.DecompressNormal(br.ReadInt32()));
                    this.Tangents.Add(ParsedModel.DecompressNormal(br.ReadInt32()));
                }

                int lightmapuvchunk = -1;
                for (int x = normalchunk + 1; x < RawDataChunkInfo.Length; x++)
                {
                    if (RawDataChunkInfo[x].ChunkSize == 3)
                    {
                        lightmapuvchunk = x;
                        break;
                    }
                }

                if (lightmapuvchunk == -1) return;
                RawDataChunkInfo[lightmapuvchunk].ChunkSize = 4;
                for (int x = 0; x < this.VerticeCount; x++)
                {
                    br.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[lightmapuvchunk].Offset + (this.RawDataChunkInfo[lightmapuvchunk].ChunkSize * x);

                    //if (u < 0) u += 1;
                    //if (v < 0) v += 1;
                    this.LightMapUVs.Add(new Vector2(DecompressVertice(Convert.ToSingle(br.ReadInt16()), -1, 1),
                        DecompressVertice(Convert.ToSingle(br.ReadInt16()), -1, 1)));
                }
            }

            public class BSPPermutationBoundingBoxContainer
            {
                public float
                MinX = 0f, MaxX = 0f,
                MinY = 0f, MaxY = 0f,
                MinZ = 0f, MaxZ = 0f,
                MinU = 0f, MaxU = 0f,
                MinV = 0f, MaxV = 0f;

                public long offset;
                public Matrix mat;

                public BSPPermutationBoundingBoxContainer(ref BinaryReader BR, ref Mesh box)
                {
                    offset = BR.BaseStream.Position;
                    MinX = BR.ReadSingle(); MaxX = BR.ReadSingle();
                    MinY = BR.ReadSingle(); MaxY = BR.ReadSingle();
                    MinZ = BR.ReadSingle(); MaxZ = BR.ReadSingle();
                    MinU = BR.ReadSingle(); MaxU = BR.ReadSingle();
                    MinV = BR.ReadSingle(); MaxV = BR.ReadSingle();

                    mat = Matrix.Translation(MinX, MinY, MinZ);
                    //MessageBox.Show((MinX+MaxX).ToString());
                    //
                    box = Mesh.Box(Map.dev, MinX + MaxX, MinY + MaxY, MinZ + MaxZ);
                }
            }
        }

        public class PermutationPlacement
        {
           public short sceneryIndex;
            public Matrix mat = Matrix.Identity;
            public float scale, x, y, z;

            public long startInMeta;
            public Mesh mesh;
            public bool IsSelected = false;

            public PermutationPlacement(ref BinaryReader br)
            {
                startInMeta = br.BaseStream.Position;

                /*byte[] tmpBuff = br.ReadBytes(0x58);
                scale = BitConverter.ToSingle(tmpBuff, 0x0);

                mat = Matrix.Identity;
                mat.M11 = BitConverter.ToSingle(tmpBuff, 0x4);
                mat.M12 = BitConverter.ToSingle(tmpBuff, 0x8);
                mat.M13 = BitConverter.ToSingle(tmpBuff, 0xC);
                mat.M14 = 0;

                mat.M21 = BitConverter.ToSingle(tmpBuff, 0x10);
                mat.M22 = BitConverter.ToSingle(tmpBuff, 0x14);
                mat.M23 = BitConverter.ToSingle(tmpBuff, 0x18);
                mat.M24 = 0;

                mat.M31 = BitConverter.ToSingle(tmpBuff, 0x1C);
                mat.M32 = BitConverter.ToSingle(tmpBuff, 0x20);
                mat.M33 = BitConverter.ToSingle(tmpBuff, 0x24);
                mat.M34 = 0;

                mat.M41 = BitConverter.ToSingle(tmpBuff, 0x28);
                mat.M42 = BitConverter.ToSingle(tmpBuff, 0x2C);
                mat.M43 = BitConverter.ToSingle(tmpBuff, 0x30);
                mat.M44 = 1;

                sceneryIndex = BitConverter.ToInt16(tmpBuff, 0x34); //dec=52
                tmpBuff = null;*/

                scale = br.ReadSingle();

                mat = Matrix.Identity;
                mat.M11 = br.ReadSingle();
                mat.M12 = br.ReadSingle();
                mat.M13 = br.ReadSingle();
                mat.M14 = 0;

                mat.M21 = br.ReadSingle();
                mat.M22 = br.ReadSingle();
                mat.M23 = br.ReadSingle();
                mat.M24 = 0;

                mat.M31 = br.ReadSingle();
                mat.M32 = br.ReadSingle();
                mat.M33 = br.ReadSingle();
                mat.M34 = 0;

                mat.M41 = br.ReadSingle();
                mat.M42 = br.ReadSingle();
                mat.M43 = br.ReadSingle();
                mat.M44 = 1;

                sceneryIndex = br.ReadInt16(); ; //dec=52

                x = mat.M41;
                y = mat.M42;
                z = mat.M43;
            }
        }

        public void LoadPermutations(ref Meta meta)
        {
            BinaryReader br = new BinaryReader(meta.MS);
            br.BaseStream.Position = 328;

            int tempc = br.ReadInt32(), 
                tempr = br.ReadInt32() - meta.magic - meta.offset;

            BSPPermutationRawDataMetaChunks = new BSPPermutationRawDataMetaChunk[tempc];

            for (int x = 0; x < tempc; x++)
                BSPPermutationRawDataMetaChunks[x] = new 
                    BSPPermutationRawDataMetaChunk(tempr + (x * 200), x, ref meta);

            br.BaseStream.Position = 336;
            tempc = br.ReadInt32();
            tempr = br.ReadInt32() - meta.magic - meta.offset;

            PermutationInfo = new PermutationPlacement[tempc];
            for (int x = 0; x < tempc; x++)
            {
                br.BaseStream.Position = tempr + (x * 88);
                //MessageBox.Show((meta.offset + br.BaseStream.Position).ToString());
                PermutationInfo[x] = new PermutationPlacement(ref br);

            }
        }

        public void LoadLightmaps(int mapnum)
        {
            if (Map.mp[mapnum].bsp.sbsp[this.BspNumber].LightMap_TagNumber == -1)
            {
                LightMapBitmap = new System.Drawing.Bitmap[this.BSPRawDataMetaChunks.Length];
                SceneryLightMapBitmap = new System.Drawing.Bitmap[this.PermutationInfo.Length];
                return;
            }

            //open the map in a reader
            BinaryReader br = new BinaryReader(new FileStream(Map.mp[mapnum].mapLoc,
                FileMode.Open, FileAccess.Read, FileShare.ReadWrite));

            Meta meta = new Meta(Map.mp[mapnum].bsp.sbsp[this.BspNumber].LightMap_TagNumber);
            meta.ReadMetaFromMap(false, true, false, mapnum);

            ParsedBitmap pb = new ParsedBitmap(ref meta, mapnum);
            LightMapBitmap = new System.Drawing.Bitmap[this.BSPRawDataMetaChunks.Length];
            for (int x = 0; x < this.BSPRawDataMetaChunks.Length; x++)
            {
                int visualchunk = Map.mp[mapnum].bsp.sbsp[this.BspNumber].VisualChunk_Bitmap_Index[x];
                if (visualchunk == -1)
                {
                    LightMapBitmap[x] = null;
                    continue;
                }
                LightMapBitmap[x] = pb.FindChunkAndDecode(visualchunk, 0, ref meta, x, this.BspNumber, mapnum);
            }

            SceneryLightMapBitmap = new System.Drawing.Bitmap[this.PermutationInfo.Length];
            for (int x = 0; x < this.PermutationInfo.Length; x++)
            {
                int visualchunk = Map.mp[mapnum].bsp.sbsp[this.BspNumber].SceneryChunk_Bitmap_Index[x];
                if (visualchunk == -1)
                {
                    SceneryLightMapBitmap[x] = null;
                    continue;
                }
                SceneryLightMapBitmap[x] = pb.FindChunkAndDecode(visualchunk, 0, ref meta, -x - 1, this.BspNumber, mapnum);
            }
            br.Close();
        }

        public void ExtractModel(string path)
        {
            if (path[path.Length - 1] != '\\') path += "\\";

            // ExportTextures
            string texturepath = path + "Textures\\";
            Directory.CreateDirectory(texturepath);

            List<string> names = new List<string>();
            string mtllib = this.Name + ".mtl";

            StreamWriter sw = new StreamWriter(new FileStream(path + mtllib, FileMode.Create));

            for (int x = 0; x < this.Shaders.Shader.Length; x++)
            {
                string temps = this.Shaders.Shader[x].shaderName.Substring
                    (this.Shaders.Shader[x].shaderName.LastIndexOf('\\') + 1);

                temps = temps.Replace(' ', '_');
                names.Add(temps);
                this.Shaders.Shader[x].MakeTextures();

                try
                {
                    Microsoft.DirectX.Direct3D.TextureLoader.Save
                       (texturepath + temps + ".dds", Microsoft.DirectX.Direct3D.ImageFileFormat.Dds,
                       this.Shaders.Shader[x].MainTexture);
                }
                catch
                { }

                if (x > 0)
                    sw.WriteLine("");

                sw.WriteLine("newmtl " + temps);
                sw.WriteLine("Ka 1.000000 1.000000 1.000000");
                sw.WriteLine("Kd 1.000000 1.000000 1.000000");
                sw.WriteLine("Ks 1.000000 1.000000 1.000000");
                sw.WriteLine("Ns 0.000000");
                sw.WriteLine("map_Kd .\\Textures\\" + temps + ".dds");
            }
            sw.Close();

            // ExportBSPMeshes
            for (int x = 0; x < this.BSPRawDataMetaChunks.Length; x++)
            {
                if (this.BSPRawDataMetaChunks[x].VerticeCount == 0) continue;

                sw = new StreamWriter(new FileStream(path + this.Name + "[" + x.ToString() + "].obj", FileMode.Create));

                sw.WriteLine("# ------------------------------------");
                sw.WriteLine("# Halo 2 BSP Mesh - Extracted with Entity");
                sw.WriteLine("# ------------------------------------");
                sw.WriteLine("mtllib " + mtllib);

                for (int y = 0; y < this.BSPRawDataMetaChunks[x].VerticeCount; y++)
                {
                    sw.WriteLine("v " + this.BSPRawDataMetaChunks[x].Vertices[y].X.ToString()
                                 + " " + this.BSPRawDataMetaChunks[x].Vertices[y].Y.ToString()
                                 + " " + this.BSPRawDataMetaChunks[x].Vertices[y].Z.ToString());
                }
                sw.WriteLine("# " + this.BSPRawDataMetaChunks[x].Vertices.Count.ToString() + " vertices");

                for (int y = 0; y < this.BSPRawDataMetaChunks[x].VerticeCount; y++)
                {
                    sw.WriteLine("vt " + this.BSPRawDataMetaChunks[x].UVs[y].X.ToString()
                                 + " " + this.BSPRawDataMetaChunks[x].UVs[y].Y.ToString());
                }
                sw.WriteLine("# " + this.BSPRawDataMetaChunks[x].Vertices.Count.ToString() + " texture vertices");

                for (int y = 0; y < this.BSPRawDataMetaChunks[x].VerticeCount; y++)
                {
                    sw.WriteLine("vn " + this.BSPRawDataMetaChunks[x].Normals[y].X
                                 + " " + this.BSPRawDataMetaChunks[x].Normals[y].Y
                                 + " " + this.BSPRawDataMetaChunks[x].Normals[y].Z);
                }
                sw.WriteLine("# " + this.BSPRawDataMetaChunks[x].Vertices.Count.ToString() + " normals");

                for (int y = 0; y < this.BSPRawDataMetaChunks[x].SubMeshInfo.Length; y++)
                {
                    sw.WriteLine("g 0." + y.ToString());
                    //sw.WriteLine("s 0." + y.ToString());
                    sw.WriteLine("usemtl  " + names[this.BSPRawDataMetaChunks[x].SubMeshInfo[y].ShaderNumber]);

                    short[] shortsArray = new short[100000];
                    int s = 0;
                    //if (this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount == (this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount/3)*3)
                    if (this.BSPRawDataMetaChunks[x].FaceCount * 3 != this.BSPRawDataMetaChunks[x].IndiceCount)
                    {
                        int m = this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart;

                        bool dir = false;
                        short tempx, tempy, tempz;

                        do
                        {
                            //if (mode.EndOfIndices[x][j]>m+2){break;}
                            tempx = this.BSPRawDataMetaChunks[x].Indices[m];
                            tempy = this.BSPRawDataMetaChunks[x].Indices[m + 1];
                            tempz = this.BSPRawDataMetaChunks[x].Indices[m + 2];

                            if (tempx != tempy && tempx != tempz && tempy != tempz)
                            {
                                if (!dir)
                                {
                                    shortsArray[s] = tempx;
                                    shortsArray[s + 1] = tempy;
                                    shortsArray[s + 2] = tempz;
                                    s += 3;

                                    dir = true;
                                }
                                else
                                {
                                    shortsArray[s] = tempx;
                                    shortsArray[s + 1] = tempz;
                                    shortsArray[s + 2] = tempy;
                                    s += 3;
                                    dir = false;
                                }
                                m += 1;
                            }
                            else
                            {
                                if (dir) dir = false;
                                else dir = true;

                                m += 1;
                            }
                        }
                        while (m < this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart + this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount - 2);
                    }
                    else
                    {
                        //Array.Copy(this.BSPRawDataMetaChunks[x].Indices, this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart, shortsArray, 0, this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount);
                        s = this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount;
                    }

                    for (int xx = 0; xx < s; xx += 3)
                    {
                        sw.WriteLine("f " + (shortsArray[xx] + 1) + "/" + (shortsArray[xx] + 1) + "/" + (shortsArray[xx] + 1)
                                    + " " + (shortsArray[xx + 1] + 1) + "/" + (shortsArray[xx + 1] + 1) + "/" + (shortsArray[xx + 1] + 1)
                                    + " " + (shortsArray[xx + 2] + 1) + "/" + (shortsArray[xx + 2] + 1) + "/" + (shortsArray[xx + 2] + 1));
                    }
                    sw.WriteLine("# " + (s / 3) + " elements");

                }
                sw.Close();
            }

            // ExportBSPPermutationMeshes
            for (int x = 0; x < this.BSPPermutationRawDataMetaChunks.Length; x++)
            {
                if (this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo == null) continue;

                sw = new StreamWriter(new FileStream(path + this.Name + "-Permutation[" + x.ToString() + "].obj", FileMode.Create));
                sw.WriteLine("# ------------------------------------");
                sw.WriteLine("# Halo 2 BSP Permutation Mesh - Extracted with Entity");
                sw.WriteLine("# ------------------------------------");
                sw.WriteLine("mtllib " + mtllib);

                for (int y = 0; y < this.BSPPermutationRawDataMetaChunks[x].VerticeCount; y++)
                {
                    sw.WriteLine("v " + this.BSPPermutationRawDataMetaChunks[x].Vertices[y].X.ToString()
                                 + " " + this.BSPPermutationRawDataMetaChunks[x].Vertices[y].Y.ToString()
                                 + " " + this.BSPPermutationRawDataMetaChunks[x].Vertices[y].Z.ToString());
                }
                sw.WriteLine("# " + this.BSPPermutationRawDataMetaChunks[x].Vertices.Count.ToString() + " vertices");

                for (int y = 0; y < this.BSPPermutationRawDataMetaChunks[x].VerticeCount; y++)
                {
                    sw.WriteLine("vt " + this.BSPPermutationRawDataMetaChunks[x].UVs[y].X.ToString()
                                 + " " + this.BSPPermutationRawDataMetaChunks[x].UVs[y].Y.ToString());
                }
                sw.WriteLine("# " + this.BSPPermutationRawDataMetaChunks[x].Vertices.Count.ToString() + " texture vertices");

                for (int y = 0; y < this.BSPPermutationRawDataMetaChunks[x].VerticeCount; y++)
                {
                    sw.WriteLine("vn " + this.BSPPermutationRawDataMetaChunks[x].Normals[y].X
                                 + " " + this.BSPPermutationRawDataMetaChunks[x].Normals[y].Y
                                 + " " + this.BSPPermutationRawDataMetaChunks[x].Normals[y].Z);
                }
                sw.WriteLine("# " + this.BSPPermutationRawDataMetaChunks[x].Vertices.Count.ToString() + " normals");

                for (int y = 0; y < this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo.Length; y++)
                {
                    sw.WriteLine("g 0." + y.ToString());
                    //  SW.WriteLine("s 0." + y.ToString());
                    sw.WriteLine("usemtl  " + names[this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].ShaderNumber]);

                    short[] shortsArray = new short[100000];
                    int s = 0;
                    if (this.BSPPermutationRawDataMetaChunks[x].FaceCount * 3 != this.BSPPermutationRawDataMetaChunks[x].IndiceCount)
                    {
                        int m = this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart;

                        bool dir = false;
                        short tempx, tempy, tempz;

                        do
                        {
                            //if (mode.EndOfIndices[x][j]>m+2){break;}
                            tempx = this.BSPPermutationRawDataMetaChunks[x].Indices[m];
                            tempy = this.BSPPermutationRawDataMetaChunks[x].Indices[m + 1];
                            tempz = this.BSPPermutationRawDataMetaChunks[x].Indices[m + 2];

                            if (tempx != tempy && tempx != tempz && tempy != tempz)
                            {
                                if (dir == false)
                                {
                                    shortsArray[s] = tempx;
                                    shortsArray[s + 1] = tempy;
                                    shortsArray[s + 2] = tempz;
                                    s += 3;

                                    dir = true;
                                }
                                else
                                {
                                    shortsArray[s] = tempx;
                                    shortsArray[s + 1] = tempz;
                                    shortsArray[s + 2] = tempy;
                                    s += 3;
                                    dir = false;
                                }
                            }
                            else
                            {
                                if (dir) dir = false;
                                else dir = true;
                            }
                            m += 1;
                        }
                        while (m < this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart + this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount - 2);
                    }
                    else
                    {
                        //Array.Copy(this.BSPPermutationRawDataMetaChunks[x].Indices, this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart, shortsArray, 0, this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount);
                        s = this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount;
                    }

                    for (int xx = 0; xx < s; xx += 3)
                    {
                        sw.WriteLine("f " + (shortsArray[xx] + 1) + "/" + (shortsArray[xx] + 1) + "/" + (shortsArray[xx] + 1)
                                    + " " + (shortsArray[xx + 1] + 1) + "/" + (shortsArray[xx + 1] + 1) + "/" + (shortsArray[xx + 1] + 1)
                                    + " " + (shortsArray[xx + 2] + 1) + "/" + (shortsArray[xx + 2] + 1) + "/" + (shortsArray[xx + 2] + 1));
                    }
                    sw.WriteLine("# " + (s / 3) + " elements");
                }
                sw.Close();
            }

            if (this.UnknownChunks == null) return;

            // ExportUnknown
            for (int x = 0; x < this.UnknownChunks.Length; x++)
            {
                if (this.UnknownChunks[x].IndiceCount == 0)
                    continue;

                sw = new StreamWriter(new FileStream(path + this.Name + "-Unknown[" + x.ToString() + "].obj", FileMode.Create));
                sw.WriteLine("# ------------------------------------");
                sw.WriteLine("# Halo 2 BSP Unknown Mesh - Extracted with Entity");
                sw.WriteLine("# ------------------------------------");

                for (int y = 0; y < this.UnknownChunks[x].VerticeCount; y++)
                {
                    sw.WriteLine("v " + this.UnknownChunks[x].Vertices[y].X.ToString()
                                 + " " + this.UnknownChunks[x].Vertices[y].Y.ToString()
                                 + " " + this.UnknownChunks[x].Vertices[y].Z.ToString());
                }
                sw.WriteLine("# " + this.UnknownChunks[x].Vertices.Count.ToString() + " vertices");
                sw.WriteLine("g 0.0");

                short[] shortsArray = new short[100000];
                int s = 0;

                Array.Copy(this.UnknownChunks[x].Indices, 0, shortsArray, 0, UnknownChunks[x].IndiceCount);
                s = this.UnknownChunks[x].IndiceCount;

                for (int xx = 0; xx < s; xx += 3)
                {
                    sw.WriteLine("f " + (shortsArray[xx] + 1) + "/" + (shortsArray[xx] + 1) + "/" + (shortsArray[xx] + 1)
                                + " " + (shortsArray[xx + 1] + 1) + "/" + (shortsArray[xx + 1] + 1) + "/" + (shortsArray[xx + 1] + 1)
                                + " " + (shortsArray[xx + 2] + 1) + "/" + (shortsArray[xx + 2] + 1) + "/" + (shortsArray[xx + 2] + 1));
                }
                sw.WriteLine("# " + (s / 3) + " elements");

                sw.Close();
            }
        }

        public void ExtractModelAsSingleMesh(string path)
        {
            string fuck = path.Remove(path.LastIndexOf('\\') + 1);

            // ExportTextures
            string texturepath = fuck + "Textures\\";
            Directory.CreateDirectory(texturepath);

            List<string> names = new List<string>();
            string mtllib = this.Name + ".mtl";

            StreamWriter sw = new StreamWriter(new FileStream(fuck + mtllib, FileMode.Create));
            Panel p = new Panel();

            for (int x = 0; x < this.Shaders.Shader.Length; x++)
            {
                string temps = this.Shaders.Shader[x].shaderName.Substring(this.Shaders.Shader[x].shaderName.LastIndexOf('\\') + 1);
                names.Add(temps);

                this.Shaders.Shader[x].MakeTextures();
                Microsoft.DirectX.Direct3D.TextureLoader.Save(texturepath + temps + ".dds",
                    Microsoft.DirectX.Direct3D.ImageFileFormat.Dds, this.Shaders.Shader[x].MainTexture);

                if (x > 0)
                    sw.WriteLine("");

                sw.WriteLine("newmtl " + temps);
                sw.WriteLine("Ka 1.000000 1.000000 1.000000");
                sw.WriteLine("Kd 1.000000 1.000000 1.000000");
                sw.WriteLine("Ks 1.000000 1.000000 1.000000");
                sw.WriteLine("Ns 0.000000");
                sw.WriteLine("map_Kd .\\Textures\\" + temps + ".dds");
            }
            sw.Close();

            // ExportBSPMeshes
            sw = new StreamWriter(new FileStream(path, FileMode.Create));
            sw.WriteLine("# ------------------------------------");
            sw.WriteLine("# Halo 2 BSP Mesh - Extracted with Entity");
            sw.WriteLine("# ------------------------------------");
            sw.WriteLine("mtllib " + mtllib);

            int vertcount = 0;
            for (int x = 0; x < this.BSPRawDataMetaChunks.Length; x++)
            {
                for (int y = 0; y < this.BSPRawDataMetaChunks[x].VerticeCount; y++)
                {
                    sw.WriteLine("v " + this.BSPRawDataMetaChunks[x].Vertices[y].X.ToString()
                                 + " " + this.BSPRawDataMetaChunks[x].Vertices[y].Y.ToString()
                                 + " " + this.BSPRawDataMetaChunks[x].Vertices[y].Z.ToString());
                }
                sw.WriteLine("# " + this.BSPRawDataMetaChunks[x].Vertices.Count.ToString() + " vertices");

                for (int y = 0; y < this.BSPRawDataMetaChunks[x].VerticeCount; y++)
                {
                    sw.WriteLine("vt " + this.BSPRawDataMetaChunks[x].UVs[y].X.ToString()
                                 + " " + this.BSPRawDataMetaChunks[x].UVs[y].Y.ToString());
                }
                sw.WriteLine("# " + this.BSPRawDataMetaChunks[x].Vertices.Count.ToString() + " texture vertices");

                for (int y = 0; y < this.BSPRawDataMetaChunks[x].VerticeCount; y++)
                {
                    sw.WriteLine("vn " + this.BSPRawDataMetaChunks[x].Normals[y].X
                                 + " " + this.BSPRawDataMetaChunks[x].Normals[y].Y
                                 + " " + this.BSPRawDataMetaChunks[x].Normals[y].Z);
                }
                sw.WriteLine("# " + this.BSPRawDataMetaChunks[x].Vertices.Count.ToString() + " normals");

                for (int y = 0; y < this.BSPRawDataMetaChunks[x].SubMeshInfo.Length; y++)
                {
                    sw.WriteLine("g " + x.ToString() + "." + y.ToString());
                    //sw.WriteLine("s "+x.ToString()+"." + y.ToString());
                    sw.WriteLine("usemtl  " + names[this.BSPRawDataMetaChunks[x].SubMeshInfo[y].ShaderNumber]);

                    short[] shortsArray = new short[100000];
                    int s = 0;
                    if (this.BSPRawDataMetaChunks[x].FaceCount * 3 != this.BSPRawDataMetaChunks[x].IndiceCount)
                    {
                        int m = this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart;

                        bool dir = false;
                        short tempx, tempy, tempz;

                        do
                        {
                            //if (mode.EndOfIndices[x][j]>m+2){break;}
                            tempx = this.BSPRawDataMetaChunks[x].Indices[m];
                            tempy = this.BSPRawDataMetaChunks[x].Indices[m + 1];
                            tempz = this.BSPRawDataMetaChunks[x].Indices[m + 2];

                            if (tempx != tempy && tempx != tempz && tempy != tempz)
                            {
                                if (!dir)
                                {
                                    shortsArray[s] = (short)(vertcount + tempx);
                                    shortsArray[s + 1] = (short)(vertcount + tempy);
                                    shortsArray[s + 2] = (short)(vertcount + tempz);
                                    s += 3;

                                    dir = true;
                                }
                                else
                                {
                                    shortsArray[s] = (short)(vertcount + tempx);
                                    shortsArray[s + 1] = (short)(vertcount + tempz);
                                    shortsArray[s + 2] = (short)(vertcount + tempy);
                                    s += 3;
                                    dir = false;
                                }
                            }
                            else
                            {
                                if (dir) dir = false;
                                else dir = true;
                            }
                            m += 1;
                        }
                        while (m < this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart + this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount - 2);
                    }
                    else
                    {
                        for (int u = 0; u < this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount; u++)
                            shortsArray[u] = (short)(vertcount + this.BSPRawDataMetaChunks[x].Indices[this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart + u]);

                        s = this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount;
                    }

                    for (int xx = 0; xx < s; xx += 3)
                    {
                        sw.WriteLine("f " + (shortsArray[xx] + 1) + "/" + (shortsArray[xx] + 1) + "/" + (shortsArray[xx] + 1)
                                    + " " + (shortsArray[xx + 1] + 1) + "/" + (shortsArray[xx + 1] + 1) + "/" + (shortsArray[xx + 1] + 1)
                                    + " " + (shortsArray[xx + 2] + 1) + "/" + (shortsArray[xx + 2] + 1) + "/" + (shortsArray[xx + 2] + 1));
                    }
                    sw.WriteLine("# " + (s / 3) + " elements");
                }
                vertcount += this.BSPRawDataMetaChunks[x].VerticeCount;
            }

            sw.Close();
            return;

            // ExportBSPPermutationMeshes
            sw.WriteLine("# ------------------------------------");
            sw.WriteLine("# Halo 2 BSP Permutation Mesh");
            sw.WriteLine("# ------------------------------------");
            for (int x = 0; x < this.BSPPermutationRawDataMetaChunks.Length; x++)
            {
                for (int y = 0; y < this.BSPPermutationRawDataMetaChunks[x].VerticeCount; y++)
                {
                    sw.WriteLine("v " + this.BSPPermutationRawDataMetaChunks[x].Vertices[y].X.ToString()
                                + " " + this.BSPPermutationRawDataMetaChunks[x].Vertices[y].Y.ToString()
                                + " " + this.BSPPermutationRawDataMetaChunks[x].Vertices[y].Z.ToString());
                }
                sw.WriteLine("# " + this.BSPPermutationRawDataMetaChunks[x].Vertices.Count.ToString() + " vertices");

                for (int y = 0; y < this.BSPPermutationRawDataMetaChunks[x].VerticeCount; y++)
                {
                    sw.WriteLine("vt " + this.BSPPermutationRawDataMetaChunks[x].UVs[y].X.ToString()
                                 + " " + this.BSPPermutationRawDataMetaChunks[x].UVs[y].Y.ToString());
                }
                sw.WriteLine("# " + this.BSPPermutationRawDataMetaChunks[x].Vertices.Count.ToString() + " texture vertices");

                for (int y = 0; y < this.BSPPermutationRawDataMetaChunks[x].VerticeCount; y++)
                {
                    sw.WriteLine("vn " + this.BSPPermutationRawDataMetaChunks[x].Normals[y].X
                                 + " " + this.BSPPermutationRawDataMetaChunks[x].Normals[y].Y
                                 + " " + this.BSPPermutationRawDataMetaChunks[x].Normals[y].Z);
                }
                sw.WriteLine("# " + this.BSPPermutationRawDataMetaChunks[x].Vertices.Count.ToString() + " normals");

                for (int y = 0; y < this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo.Length; y++)
                {
                    sw.WriteLine("g 0." + y.ToString());
                    //   SW.WriteLine("s 0." + y.ToString());
                    sw.WriteLine("usemtl  " + names[this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].ShaderNumber]);

                    short[] shortsArray = new short[100000];
                    int s = 0;
                    if (this.BSPPermutationRawDataMetaChunks[x].FaceCount * 3 != this.BSPPermutationRawDataMetaChunks[x].IndiceCount)
                    {
                        int m = this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart;

                        bool dir = false;
                        short tempx, tempy, tempz;

                        do
                        {
                            //if (mode.EndOfIndices[x][j]>m+2){break;}
                            tempx = this.BSPPermutationRawDataMetaChunks[x].Indices[m];
                            tempy = this.BSPPermutationRawDataMetaChunks[x].Indices[m + 1];
                            tempz = this.BSPPermutationRawDataMetaChunks[x].Indices[m + 2];

                            if (tempx != tempy && tempx != tempz && tempy != tempz)
                            {
                                if (!dir)
                                {
                                    shortsArray[s] = (short)(vertcount + tempx);
                                    shortsArray[s + 1] = (short)(vertcount + tempy);
                                    shortsArray[s + 2] = (short)(vertcount + tempz);
                                    s += 3;

                                    dir = true;
                                }
                                else
                                {
                                    shortsArray[s] = (short)(vertcount + tempx);
                                    shortsArray[s + 1] = (short)(vertcount + tempz);
                                    shortsArray[s + 2] = (short)(vertcount + tempy);
                                    s += 3;
                                    dir = false;
                                }
                            }
                            else
                            {
                                if (dir) dir = false;
                                else dir = true;
                            }
                            m += 1;
                        }
                        while (m < this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart + this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount - 2);
                    }
                    else
                    {
                        for (int u = 0; u < this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount; u++)
                            shortsArray[u] = (short)(vertcount + this.BSPPermutationRawDataMetaChunks[x].Indices[this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart + u]);

                        s = this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount;
                    }

                    for (int xx = 0; xx < s; xx += 3)
                    {
                        sw.WriteLine("f " + (shortsArray[xx] + 1) + "/" + (shortsArray[xx] + 1) + "/" + (shortsArray[xx] + 1)
                                    + " " + (shortsArray[xx + 1] + 1) + "/" + (shortsArray[xx + 1] + 1) + "/" + (shortsArray[xx + 1] + 1)
                                    + " " + (shortsArray[xx + 2] + 1) + "/" + (shortsArray[xx + 2] + 1) + "/" + (shortsArray[xx + 2] + 1));
                    }
                    sw.WriteLine("# " + (s / 3) + " elements");
                    vertcount += this.BSPPermutationRawDataMetaChunks[x].VerticeCount;
                }
            }
            sw.Close();
        }

        public class BSPCollision
        {
            public Vector4[] Planes;
            public int PlaneReflexiveCount;
            public int PlaneReflexiveOffset;
            public int PlaneReflexiveTranslation;
            public Vector3[] Vertices;
            public ushort[] Faces;
            public int FaceReflexiveCount;
            public int FaceReflexiveOffset;
            public int FaceReflexiveTranslation;
            public int VerticeReflexiveCount;
            public int VerticeReflexiveOffset;
            public int VerticeReflexiveTranslation;
            public int SurfaceReflexiveCount;
            public int SurfaceReflexiveTranslation;

            public BSPCollision(Meta meta)
            {
                BinaryReader br = new BinaryReader(meta.MS);
                br.BaseStream.Position = 36;

                int tempc = br.ReadInt32();
                int tempr = br.ReadInt32() - meta.magic - meta.offset;
                int tempblah = tempr + 44;
                SurfaceReflexiveCount = br.ReadInt32();
                SurfaceReflexiveTranslation = br.ReadInt32() - meta.magic - meta.offset;

                //planes
                PlaneReflexiveOffset = tempr + 8;
                br.BaseStream.Position = PlaneReflexiveOffset;

                PlaneReflexiveCount = br.ReadInt32();
                PlaneReflexiveTranslation = br.ReadInt32() - meta.magic - meta.offset;
                Planes = new Vector4[PlaneReflexiveCount];
                for (int x = 0; x < FaceReflexiveCount; x++)
                {
                    Planes[x].X = br.ReadSingle();
                    Planes[x].Y = br.ReadSingle();
                    Planes[x].Z = br.ReadSingle();
                    Planes[x].W = br.ReadSingle();
                }

                //faces
                FaceReflexiveOffset = tempr + 48;
                br.BaseStream.Position = FaceReflexiveOffset;
                FaceReflexiveCount = br.ReadInt32();
                FaceReflexiveTranslation = br.ReadInt32() - meta.magic - meta.offset;

                Faces = new ushort[FaceReflexiveCount * 3];
                for (int x = 0; x < FaceReflexiveCount; x++)
                {
                    br.BaseStream.Position = FaceReflexiveTranslation + (x * 12);
                    Faces[(x * 3)] = br.ReadUInt16();
                    Faces[(x * 3) + 1] = br.ReadUInt16();
                    Faces[(x * 3) + 2] = Faces[(x * 3) + 1];
                }

                //vertices
                VerticeReflexiveOffset = tempr + 56;
                br.BaseStream.Position = VerticeReflexiveOffset;
                VerticeReflexiveCount = br.ReadInt32();
                VerticeReflexiveTranslation = br.ReadInt32() - meta.magic - meta.offset;
                Vertices = new Vector3[VerticeReflexiveCount];
                for (int x = 0; x < VerticeReflexiveCount; x++)
                {
                    br.BaseStream.Position = VerticeReflexiveTranslation + (x * 16);
                    Vertices[x].X = br.ReadSingle();
                    Vertices[x].Y = br.ReadSingle();
                    Vertices[x].Z = br.ReadSingle();
                }
            }

            public void ExtractCollisonMesh(string FilePath)
            {
                StreamWriter sw = new StreamWriter(new FileStream(FilePath, FileMode.Create));
                sw.WriteLine("# ------------------------------------");
                sw.WriteLine("# Halo 2 BSP Collision Mesh - Extracted with Entity");
                sw.WriteLine("# ------------------------------------");
                for (int x = 0; x < Vertices.Length; x++)
                {
                    sw.WriteLine("v " + Vertices[x].X.ToString() + " "
                                        + Vertices[x].Y.ToString() + " "
                                        + Vertices[x].Z.ToString());
                }
                for (int x = 0; x < Faces.Length; x += 3)
                {
                    sw.WriteLine("f " + (Faces[x] + 1) + " "
                                        + (Faces[x + 1] + 1) + " "
                                        + (Faces[x + 2] + 1));
                }
                sw.Close();
            }

            public void InjectCollisonMesh(string FilePath, Meta meta, int mapnum)
            {
                StreamReader sr = new StreamReader(new FileStream(FilePath, FileMode.Open));
                string temps = "";
                int verticecount = 0, facecount = 0;
                List<Vector3> vertices = new List<Vector3>();
                List<ushort> faces = new List<ushort>();
                do
                {
                    temps = sr.ReadLine();
                    if (temps == null) continue;
                    string[] tempstrings = temps.Split(',', ' ');
                    switch (tempstrings[0])
                    {
                        case "v":
                            vertices.Add(new Vector3(float.Parse(tempstrings[1]),
                                float.Parse(tempstrings[2]), float.Parse(tempstrings[3])));
                            verticecount++;
                            break;

                        case "f":
                            string[]
                            split1 = tempstrings[1].Split('/'),
                            split2 = tempstrings[2].Split('/'),
                            split3 = tempstrings[3].Split('/');
                            ushort
                            temp1 = ushort.Parse(split1[0]),
                            temp2 = ushort.Parse(split2[0]),
                            temp3 = ushort.Parse(split3[0]);
                            temp1--;
                            temp2--;
                            temp3--;
                            faces.Add(temp1);
                            faces.Add(temp2);
                            faces.Add(temp3);
                            facecount += 3;
                            break;
                    }
                }
                while (temps != null);

                sr.Close();

                if (this.Vertices.Length >= verticecount && this.Faces.Length >= facecount)
                {
                    BinaryWriter bw = new BinaryWriter(new FileStream(Map.mp[mapnum].mapLoc,
                                        FileMode.Open, FileAccess.Write, FileShare.ReadWrite));

                    bw.BaseStream.Position = meta.offset + this.VerticeReflexiveOffset;
                    bw.Write(verticecount);
                    for (int x = 0; x < verticecount; x++)
                    {
                        bw.BaseStream.Position = meta.offset + this.VerticeReflexiveTranslation + (x * 16);
                        bw.Write(vertices[x].X);
                        bw.Write(vertices[x].Y);
                        bw.Write(vertices[x].Z);
                    }

                    bw.BaseStream.Position = meta.offset + this.FaceReflexiveOffset;
                    int tempint = (int)(facecount / 3);
                    bw.Write(tempint);

                    for (int x = 0; x < facecount; x += 3)
                    {
                        int fuck = (x / 3);
                        bw.BaseStream.Position = meta.offset + this.FaceReflexiveTranslation + (fuck * 12);
                        bw.Write(faces[x]);
                        bw.Write(faces[x + 1]);
                    }
                    bw.Close();
                }
            }
        }

        public class BSPShaderContainer : ParsedModel.ShaderContainer
        {
            public BSPShaderContainer()
            {
            }
            public BSPShaderContainer(BSPModel bsp, int mapnum)
                : base()
            {
                Meta meta = Map.mp[mapnum].meta;
                BinaryReader BR = new BinaryReader(meta.MS);
                BR.BaseStream.Position = 180;
                int tempc = BR.ReadInt32(), tempr = BR.ReadInt32() - meta.magic - meta.offset;

                this.Shader = new ShaderInfo[tempc];
                for (int x = 0; x < tempc; x++)
                {
                    BR.BaseStream.Position = tempr + (x * 32) + 12;
                    int tempint = Meta.FindMetaByID(BR.ReadInt32(), Map.mp[mapnum].tags.tagInfo);
                    this.Shader[x] = new ShaderInfo(tempint, mapnum);

                }
            }
        }

        public class BSPDisplayedInfo : ParsedModel.DisplayedInfo
        {
            // private Mesh[] mesh;
            private VertexBuffer[] permvertexBuffer = null;
            private IndexBuffer[] permindexBuffer = null;
            Matrix watermatrix = Matrix.Identity;

            public static void FindDisplayedPieces()
            { }

            public static void LoadDirectXTexturesAndBuffers(ref BSPModel bsp)
            {
                BSPDisplayedInfo.CreateVertexBuffers(ref bsp);
                BSPDisplayedInfo.CreateIndexBuffers(ref bsp);
                BSPDisplayedInfo.LoadShaderTextures(ref bsp);
                BSPDisplayedInfo.LoadLightmapTextures(ref bsp);
                BSPDisplayedInfo.MakeScenMeshes(ref bsp);
            }

            public static void MakeScenMeshes(ref BSPModel bsp)
            {
                for (int x = 0; x < bsp.PermutationInfo.Length; x++)
                {
                    int rawindex = bsp.PermutationInfo[x].sceneryIndex;
                    if (bsp.BSPPermutationRawDataMetaChunks[rawindex].RawDataChunkInfo.Length == 0)
                        continue;

                    bsp.PermutationInfo[x].mesh = MakeMesh(
                        bsp.BSPPermutationRawDataMetaChunks[rawindex].Vertices,
                        bsp.BSPPermutationRawDataMetaChunks[rawindex].Indices,
                        bsp.BSPPermutationRawDataMetaChunks[rawindex].UVs);
                }
            }

            public static void LoadLightmapTextures(ref BSPModel bsp)
            {
                if (bsp.renderLightmap)
                {
                    bsp.LightMapTexture = new Texture[bsp.BSPRawDataMetaChunks.Length];
                    for (int x = 0; x < bsp.LightMapBitmap.Length; x++)
                        bsp.LightMapTexture[x] = ShaderInfo.CreateTexture(bsp.LightMapBitmap[x]);
                }

                bsp.SceneryLightMapTexture = new Texture[bsp.SceneryLightMapBitmap.Length];
                for (int x = 0; x < bsp.SceneryLightMapBitmap.Length; x++)
                    bsp.SceneryLightMapTexture[x] = ShaderInfo.CreateTexture(bsp.SceneryLightMapBitmap[x]);
            }

            public static void LoadShaderTextures(ref BSPModel bsp)
            {
                for (int x = 0; x < bsp.Shaders.Shader.Length; x++)
                {
                    bsp.Shaders.Shader[x].MakeTextures();
                    bsp.Shaders.Shader[x].KillBitmaps();
                }
            }

            public static void CreateIndexBuffers(ref BSPModel bsp)
            {
                bsp.Display.indexBuffer = new IndexBuffer[bsp.BSPRawDataMetaChunks.Length];
                for (int x = 0; x < bsp.BSPRawDataMetaChunks.Length; x++)
                {
                    if (bsp.BSPRawDataMetaChunks[x].RawDataChunkInfo.Length == 0)
                        continue;

                    bsp.Display.indexBuffer[x] = new IndexBuffer(
                        typeof(short), bsp.BSPRawDataMetaChunks[x].IndiceCount,
                        Map.dev,Usage.WriteOnly, Pool.Managed);

                    IndexBuffer ib = (IndexBuffer)bsp.Display.indexBuffer[x];
                    ib.SetData(bsp.BSPRawDataMetaChunks[x].Indices.ToArray(), 0, LockFlags.None);
                    ib.Unlock();
                }

                bsp.Display.permindexBuffer = new IndexBuffer[bsp.BSPPermutationRawDataMetaChunks.Length];
                for (int x = 0; x < bsp.BSPPermutationRawDataMetaChunks.Length; x++)
                {
                    if (bsp.BSPPermutationRawDataMetaChunks[x].RawDataChunkInfo.Length == 0)
                        continue;

                    bsp.Display.permindexBuffer[x] = new IndexBuffer(
                        typeof(short), bsp.BSPPermutationRawDataMetaChunks[x].IndiceCount,
                        Map.dev,Usage.WriteOnly, Pool.Managed);

                    IndexBuffer ib = (IndexBuffer)bsp.Display.permindexBuffer[x];
                    ib.SetData(bsp.BSPPermutationRawDataMetaChunks[x].Indices.ToArray(), 0, LockFlags.None);
                    ib.Unlock();
                }
            }

            public static void CreateVertexBuffers(ref BSPModel bsp)
            {
                bsp.Display.vertexBuffer = new VertexBuffer[bsp.BSPRawDataMetaChunks.Length];
                for (int x = 0; x < bsp.BSPRawDataMetaChunks.Length; x++)
                {
                    int rawindex = x;
                    if (bsp.BSPRawDataMetaChunks[rawindex].RawDataChunkInfo.Length == 0)
                        continue;

                    bsp.Display.vertexBuffer[rawindex] = new VertexBuffer(
                        typeof(HaloBSPVertex), bsp.BSPRawDataMetaChunks[rawindex].VerticeCount,
                        Map.dev,Usage.WriteOnly, HaloBSPVertex.FVF, Pool.Managed);

                    // Lock the buffer (which will return our structs)
                    HaloBSPVertex[] verts = (HaloBSPVertex[])bsp.Display.vertexBuffer[rawindex].Lock(0, 0);
                    for (int i = 0; i < bsp.BSPRawDataMetaChunks[rawindex].VerticeCount; i++)
                    {
                        verts[i].Position = new Vector3(
                            bsp.BSPRawDataMetaChunks[rawindex].Vertices[i].X,
                            bsp.BSPRawDataMetaChunks[rawindex].Vertices[i].Y,
                            bsp.BSPRawDataMetaChunks[rawindex].Vertices[i].Z);

                        verts[i].Tu0 = bsp.BSPRawDataMetaChunks[rawindex].UVs[i].X;
                        verts[i].Tv0 = bsp.BSPRawDataMetaChunks[rawindex].UVs[i].Y;
                        verts[i].Normal = bsp.BSPRawDataMetaChunks[rawindex].Normals[i];

                        //verts[i].specular = 1;
                        verts[i].diffuse = 1;

                        verts[i].Tu1 = verts[i].Tu0;
                        verts[i].Tv1 = verts[i].Tv0;
                        if (bsp.BSPRawDataMetaChunks[rawindex].LightMapUVs.Count != 0)
                        {
                            verts[i].Tu2 = bsp.BSPRawDataMetaChunks[rawindex].LightMapUVs[i].X;
                            verts[i].Tv2 = bsp.BSPRawDataMetaChunks[rawindex].LightMapUVs[i].Y;
                        }
                        else
                        {
                            verts[i].Tu2 = verts[i].Tu0;
                            verts[i].Tv2 = verts[i].Tv0;
                        }
                        verts[i].Tu3 = verts[i].Tu0;
                        verts[i].Tv3 = verts[i].Tv0;
                    }
                    bsp.Display.vertexBuffer[rawindex].Unlock();
                }

                //scenery
                bsp.Display.permvertexBuffer = new VertexBuffer[bsp.BSPPermutationRawDataMetaChunks.Length];
                for (int x = 0; x < bsp.BSPPermutationRawDataMetaChunks.Length; x++)
                {
                    int rawindex = x;
                    if (bsp.BSPPermutationRawDataMetaChunks[rawindex].RawDataChunkInfo.Length == 0)
                        continue;

                    bsp.Display.permvertexBuffer[rawindex] = new VertexBuffer(
                        typeof(HaloBSPVertex), bsp.BSPPermutationRawDataMetaChunks[rawindex].VerticeCount,
                        Map.dev,Usage.WriteOnly, HaloBSPVertex.FVF, Pool.Managed);

                    HaloBSPVertex[] verts = (HaloBSPVertex[])bsp.Display.permvertexBuffer[rawindex].Lock(0, 0);
                    for (int i = 0; i < bsp.BSPPermutationRawDataMetaChunks[rawindex].VerticeCount; i++)
                    {
                        verts[i].Position = new Vector3(
                            bsp.BSPPermutationRawDataMetaChunks[rawindex].Vertices[i].X,
                            bsp.BSPPermutationRawDataMetaChunks[rawindex].Vertices[i].Y,
                        bsp.BSPPermutationRawDataMetaChunks[rawindex].Vertices[i].Z);

                        verts[i].Tu0 = bsp.BSPPermutationRawDataMetaChunks[rawindex].UVs[i].X;
                        verts[i].Tv0 = bsp.BSPPermutationRawDataMetaChunks[rawindex].UVs[i].Y;
                        verts[i].Normal = bsp.BSPPermutationRawDataMetaChunks[rawindex].Normals[i];

                        //verts[i].specular = 1;
                        verts[i].diffuse = 1;

                        verts[i].Tu1 = verts[i].Tu0;
                        verts[i].Tv1 = verts[i].Tv0;
                        if (bsp.BSPPermutationRawDataMetaChunks[rawindex].LightMapUVs.Count != 0)
                        {
                            verts[i].Tu2 = bsp.BSPPermutationRawDataMetaChunks[rawindex].LightMapUVs[i].X;
                            verts[i].Tv2 = bsp.BSPPermutationRawDataMetaChunks[rawindex].LightMapUVs[i].Y;
                        }
                        else
                        {
                            verts[i].Tu2 = verts[i].Tu0;
                            verts[i].Tv2 = verts[i].Tv0;
                        }
                        verts[i].Tu3 = verts[i].Tu0;
                        verts[i].Tv3 = verts[i].Tv0;

                    }
                    bsp.Display.permvertexBuffer[rawindex].Unlock();
                }
            }

            public static void ResetTextureStates()
            {
                for (int x = 0; x < 4; x++)
                {
                    Map.dev.SamplerState[x].AddressU =
                    Map.dev.SamplerState[x].AddressV = TextureAddress.Wrap;
                }

                Map.dev.Transform.Texture0 = Matrix.Identity;
                Map.dev.TextureState[0].TextureTransform = TextureTransform.Disable;
                Map.dev.TextureState[0].TextureCoordinateIndex = 0;
                Map.dev.TextureState[0].ColorOperation = TextureOperation.Disable;
                Map.dev.SetTexture(0, null);

                Map.dev.Transform.Texture1 = Matrix.Identity;
                Map.dev.TextureState[1].TextureTransform = TextureTransform.Disable;
                Map.dev.TextureState[1].TextureCoordinateIndex = 1;
                Map.dev.TextureState[1].ColorOperation = TextureOperation.Disable;
                Map.dev.SetTexture(1, null);

                Map.dev.Transform.Texture2 = Matrix.Identity;
                Map.dev.TextureState[2].TextureTransform = TextureTransform.Disable;
                Map.dev.TextureState[2].TextureCoordinateIndex = 2;
                Map.dev.TextureState[2].ColorOperation = TextureOperation.Disable;
                Map.dev.SetTexture(2, null);

                Map.dev.Transform.Texture3 = Matrix.Identity;
                Map.dev.TextureState[3].TextureTransform = TextureTransform.Disable;
                Map.dev.TextureState[3].TextureCoordinateIndex = 3;
                Map.dev.TextureState[3].ColorOperation = TextureOperation.Disable;
                Map.dev.SetTexture(3, null);
            }

            public static void Draw(ref BSPModel bsp, FillMode fm,
                ref Camera cam)//, DXShader shaderx)
            {
                #region BSPMainMesh
                int count = 0;
                for (int x = 0; x < bsp.BSPRawDataMetaChunks.Length; x++)
                {
                    int rawindex = x;
                    if (bsp.BSPRawDataMetaChunks[rawindex].RawDataChunkInfo.Length == 0)
                        continue;

                    Map.dev.SetStreamSource(0, bsp.Display.vertexBuffer[rawindex], 0);
                    Map.dev.VertexFormat = HaloBSPVertex.FVF;
                    Map.dev.Indices = bsp.Display.indexBuffer[rawindex];

                    for (int y = 0; y < bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo.Length; y++)
                    {
                        ResetTextureStates();
                        int tempshade = bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[y].ShaderNumber;

                        // AlphaBlending
                        SetAlphaBlending(bsp.Shaders.Shader[tempshade].Alpha);

                        if (fm == FillMode.Solid)
                        {
                            if (bsp.Shaders.Shader[tempshade].MainName.IndexOf("water!") != -1)
                            {
                                for (int u = 0; u < bsp.Shaders.Shader[tempshade].BitmapTextures.Length; u++)
                                {
                                    if (bsp.Shaders.Shader[tempshade].BitmapNames[u].IndexOf("reflection") != -1 || 
                                        bsp.Shaders.Shader[tempshade].BitmapNames[u].IndexOf("mask") != -1)
                                    {
                                        Map.dev.SetTexture(0, bsp.Shaders.Shader[tempshade].BitmapTextures[u]);
                                        if (bsp.Shaders.Shader[tempshade].MainName.IndexOf("mip") != -1)
                                        {
                                            Map.dev.SetTexture(0, bsp.Shaders.Shader[tempshade].BitmapTextures[u]);
                                            Map.dev.SetTexture(1, bsp.Shaders.Shader[tempshade].BitmapTextures[u]);
                                        }
                                        else Map.dev.SetTexture(1, bsp.Shaders.Shader[tempshade].MainTexture);

                                        Map.dev.TextureState[0].ColorOperation = TextureOperation.BumpEnvironmentMap;
                                        Map.dev.TextureState[0].ColorArgument1 = TextureArgument.TextureColor;
                                        Map.dev.TextureState[0].ColorArgument2 = TextureArgument.Current;

                                        float r = 0.07F;
                                        Map.dev.TextureState[0].BumpEnvironmentMaterial00 = r * (float)Math.Cos((Environment.TickCount / 4) * 11.0F);
                                        Map.dev.TextureState[0].BumpEnvironmentMaterial01 = -r * (float)Math.Sin((Environment.TickCount / 4) * 11.0F);
                                        Map.dev.TextureState[0].BumpEnvironmentMaterial10 = r * (float)Math.Sin((Environment.TickCount / 4) * 11.0F);
                                        Map.dev.TextureState[0].BumpEnvironmentMaterial11 = r * (float)Math.Cos((Environment.TickCount / 4) * 11.0F);

                                        Map.dev.TextureState[1].ColorOperation = TextureOperation.Modulate;
                                        Map.dev.TextureState[1].ColorArgument1 = TextureArgument.TextureColor;
                                        Map.dev.TextureState[1].ColorArgument2 = TextureArgument.Current;

                                        if (bsp.renderLightmap && bsp.LightMapTexture[x] != null &&
                                            bsp.BSPRawDataMetaChunks[x].LightMapUVs.Count != 0)
                                        {
                                            Map.dev.SetTexture(2, bsp.LightMapTexture[x]);
                                            Map.dev.TextureState[2].ColorOperation = TextureOperation.Modulate2X;
                                            Map.dev.TextureState[2].ColorArgument1 = TextureArgument.TextureColor;
                                            Map.dev.TextureState[2].ColorArgument2 = TextureArgument.Current;
                                        }
                                        break;
                                    }
                                    if (u == bsp.Shaders.Shader[tempshade].BitmapTextures.Length - 1)
                                    {
                                        Map.dev.SetTexture(0, bsp.Shaders.Shader[tempshade].MainTexture);
                                        Map.dev.TextureState[0].ColorOperation = TextureOperation.Modulate;
                                        Map.dev.TextureState[0].ColorArgument1 = TextureArgument.TextureColor;
                                        Map.dev.TextureState[0].ColorArgument2 = TextureArgument.Current;
                                    }
                                }
                            }
                            else if (bsp.Shaders.Shader[tempshade].MainName.IndexOf("ground!") != -1)
                            {
                                if (bsp.Shaders.Shader[tempshade].MainName.IndexOf("unwrapped") != -1)
                                {
                                     Map.dev.SamplerState[0].AddressU = TextureAddress.Clamp;
                                     Map.dev.SamplerState[0].AddressV = TextureAddress.Clamp;
                                }

                                Map.dev.SetTexture(0, bsp.Shaders.Shader[tempshade].MainTexture);
                                Map.dev.TextureState[0].ColorOperation = TextureOperation.Modulate;
                                Map.dev.TextureState[0].ColorArgument1 = TextureArgument.TextureColor;
                                Map.dev.TextureState[0].ColorArgument2 = TextureArgument.Current;
                                Map.dev.TextureState[0].AlphaOperation = TextureOperation.SelectArg1;
                                Map.dev.TextureState[0].AlphaArgument1 = TextureArgument.TextureColor;
                                
                                //  Map.dev.TextureState[0].ResultArgument = TextureArgument.Current;
                                Map.dev.TextureState[1].TextureTransform = TextureTransform.Count2;
                                Map.dev.Transform.Texture1 = Matrix.Scaling(bsp.Shaders.Shader[tempshade].primarydetailuscale, bsp.Shaders.Shader[tempshade].primarydetailvscale, bsp.Shaders.Shader[tempshade].primarydetailwscale);
                                // Map.dev.Transform.Texture1 = Matrix.Scaling(bsp.Shaders.Shader[tempshade].secondarydetailuscale, bsp.Shaders.Shader[tempshade].secondarydetailvscale, bsp.Shaders.Shader[tempshade].secondarydetailwscale);
                                
                                Map.dev.SetTexture(1, bsp.Shaders.Shader[tempshade].BitmapTextures[2]);
                                Map.dev.TextureState[1].ColorOperation = TextureOperation.Modulate2X;
                                Map.dev.TextureState[1].ColorArgument1 = TextureArgument.TextureColor;
                                Map.dev.TextureState[1].ColorArgument2 = TextureArgument.Current;
                                Map.dev.TextureState[1].AlphaOperation = TextureOperation.Add;
                                Map.dev.TextureState[1].AlphaArgument1 = TextureArgument.Current;

                                Map.dev.TextureState[2].TextureTransform = TextureTransform.Count2;
                                Map.dev.Transform.Texture2 = Matrix.Scaling(bsp.Shaders.Shader[tempshade].secondarydetailuscale, bsp.Shaders.Shader[tempshade].secondarydetailvscale, bsp.Shaders.Shader[tempshade].secondarydetailwscale);
                                Map.dev.SetTexture(2, bsp.Shaders.Shader[tempshade].BitmapTextures[3]);
                                Map.dev.TextureState[2].ColorOperation = TextureOperation.Modulate2X;
                                Map.dev.TextureState[2].ColorArgument1 = TextureArgument.TextureColor;
                                Map.dev.TextureState[2].ColorArgument2 = TextureArgument.Current;
                                Map.dev.TextureState[2].TextureCoordinateIndex = 3;

                                if (bsp.renderLightmap && bsp.LightMapTexture[x] != null && 
                                    bsp.BSPRawDataMetaChunks[x].LightMapUVs.Count != 0)
                                {
                                    //device.SetTexture(3, bsp.LightMapTexture[x]);
                                   // Map.dev.TextureState[3].ColorOperation = TextureOperation.Modulate2X;
                                   // Map.dev.TextureState[3].ColorArgument1 = TextureArgument.TextureColor;
                                    //device.TextureState[3].ColorArgument2 = TextureArgument.Current;
                                    //device.TextureState[3].TextureCoordinateIndex = 2;
                                }
                            }
                            else
                            {
                                Map.dev.SetTexture(0, bsp.Shaders.Shader[tempshade].MainTexture);
                                Map.dev.TextureState[0].ColorOperation = TextureOperation.Modulate;
                                Map.dev.TextureState[0].ColorArgument1 = TextureArgument.TextureColor;
                                Map.dev.TextureState[0].ColorArgument2 = TextureArgument.Current;
                                Map.dev.TextureState[0].AlphaOperation = TextureOperation.BlendTextureAlpha;
                                Map.dev.TextureState[0].AlphaArgument1 = TextureArgument.TextureColor;
                                Map.dev.TextureState[0].AlphaArgument2 = TextureArgument.Diffuse;
                                
                                /*
                                //if (bsp.Shaders.Shader[tempshade].microdetailName != null)
                                //{
                                   Map.dev.TextureState[1].TextureTransform = TextureTransform.Count2;
                                   Map.dev.Transform.Texture1 = Matrix.Scaling(bsp.Shaders.Shader[tempshade].microdetailuscale, bsp.Shaders.Shader[tempshade].microdetailvscale, bsp.Shaders.Shader[tempshade].microdetailwscale);
                                   Map.dev.SetTexture(1, bsp.Shaders.Shader[tempshade].microdetailTexture);
                                   Map.dev.TextureState[1].ColorOperation = TextureOperation.Modulate2X;
                                   Map.dev.TextureState[1].ColorArgument1 = TextureArgument.TextureColor;
                                   Map.dev.TextureState[1].ColorArgument2 = TextureArgument.Current;
                                   Map.dev.TextureState[1].AlphaOperation = TextureOperation.BlendCurrentAlpha;
                                   Map.dev.TextureState[1].AlphaArgument1 = TextureArgument.Current;
                                //}
                        
                                //if (bsp.Shaders.Shader[tempshade].primarydetailTexture != null)
                                //{
                                  Map.dev.TextureState[1].TextureTransform = TextureTransform.Count2;
                                  Map.dev.Transform.Texture1 = Matrix.Scaling(bsp.Shaders.Shader[tempshade].primarydetailuscale, bsp.Shaders.Shader[tempshade].primarydetailvscale, bsp.Shaders.Shader[tempshade].primarydetailwscale);
                                  Map.dev.SetTexture(1, bsp.Shaders.Shader[tempshade].primarydetailTexture);
                                  Map.dev.TextureState[1].ColorOperation = TextureOperation.Lerp;
                                  Map.dev.TextureState[1].ColorArgument1 = TextureArgument.TextureColor;
                                  Map.dev.TextureState[1].ColorArgument2 = TextureArgument.Current;
                                  Map.dev.TextureState[1].AlphaOperation = TextureOperation.Modulate;
                                  Map.dev.TextureState[1].AlphaArgument1 = TextureArgument.TextureColor;
                                //}
                           
                                //if (bsp.Shaders.Shader[tempshade].secondarydetailTexture != null)
                                //{
                                    Map.dev.TextureState[2].TextureTransform = TextureTransform.Count2;
                                    Map.dev.Transform.Texture2 = Matrix.Scaling(bsp.Shaders.Shader[tempshade].secondarydetailuscale, bsp.Shaders.Shader[tempshade].secondarydetailvscale, bsp.Shaders.Shader[tempshade].secondarydetailwscale);
                                    Map.dev.SetTexture(2, bsp.Shaders.Shader[tempshade].secondarydetailTexture);
                                    Map.dev.TextureState[2].ColorOperation = TextureOperation.Lerp;
                                    Map.dev.TextureState[2].ColorArgument1 = TextureArgument.TextureColor;
                                    Map.dev.TextureState[2].ColorArgument2 = TextureArgument.Current;
                                    Map.dev.TextureState[2].TextureCoordinateIndex = 3;
                                    //device.TextureState[2].AlphaOperation = TextureOperation.BlendCurrentAlpha;
                                    //device.TextureState[1].AlphaArgument1 = TextureArgument.Current;
                   
                                //} */

                                if (bsp.renderLightmap && bsp.LightMapTexture[x] != null && 
                                    bsp.BSPRawDataMetaChunks[x].LightMapUVs.Count != 0)
                                {
                                    Map.dev.SetTexture(1, bsp.LightMapTexture[x]);
                                    Map.dev.TextureState[1].ColorOperation = TextureOperation.Modulate2X;
                                    Map.dev.TextureState[1].ColorArgument1 = TextureArgument.TextureColor;
                                    Map.dev.TextureState[1].ColorArgument2 = TextureArgument.Current;
                                    Map.dev.TextureState[1].TextureCoordinateIndex = 2;
                                }
                            }

                            Map.dev.RenderState.CullMode = Cull.None;

                            //draw bsp
                            if (bsp.BSPRawDataMetaChunks[rawindex].FaceCount * 3 != 
                                bsp.BSPRawDataMetaChunks[rawindex].IndiceCount)
                            {
                                Map.dev.RenderState.FillMode = FillMode.Solid;
                                Material mm = new Material();
                                mm.Diffuse = System.Drawing.Color.White;
                                mm.Ambient = System.Drawing.Color.White;
                                //     if (x == 1) { mm.Diffuse = System.Drawing.Color.Red; mm.Ambient = System.Drawing.Color.Red; }
                                
                                Map.dev.Material = mm;
                                Map.dev.DrawIndexedPrimitives(
                                    PrimitiveType.TriangleStrip,
                                    0,
                                    0,
                                    bsp.BSPRawDataMetaChunks[rawindex].VerticeCount,
                                    bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[y].IndiceStart,
                                    bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[y].IndiceCount
                                 );
                            }
                            else
                            {
                                count++;

                                Map.dev.RenderState.FillMode = FillMode.Solid;
                                Material mm = new Material();
                                mm.Diffuse = System.Drawing.Color.White;
                                mm.Ambient = System.Drawing.Color.White;
                                //   if (x == 1) { mm.Diffuse = System.Drawing.Color.Red; mm.Ambient = System.Drawing.Color.Red; }
                                Map.dev.Material = mm;
                                Map.dev.DrawIndexedPrimitives(
                                    PrimitiveType.TriangleList,
                                    0,
                                    0,
                                    bsp.BSPRawDataMetaChunks[rawindex].VerticeCount,
                                    bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[y].IndiceStart,
                                    bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[y].IndiceCount / 3
                                 );
                            }
                        }
                        else //fm = wireframe/point
                        {
                            Map.dev.SetTexture(0, null);
                            Map.dev.SetTexture(1, null);

                            Material m = new Material();
                            m.Diffuse = System.Drawing.Color.White;
                            m.Ambient = System.Drawing.Color.White;
                            Map.dev.Material = m;
                            Map.dev.RenderState.FillMode = fm;
                            Map.dev.DrawIndexedPrimitives(
                                PrimitiveType.TriangleStrip,
                                0,
                                0,
                                bsp.BSPRawDataMetaChunks[rawindex].VerticeCount,
                                bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[y].IndiceStart,
                                bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[y].IndiceCount / 3);
                        }
                    }
                }
                #endregion

                //draw scenery ( BSPPermutationMesh )
                for (int x = 0; x < bsp.PermutationInfo.Length; x++)
                {
                    int rawindex = bsp.PermutationInfo[x].sceneryIndex;
                    if (bsp.BSPPermutationRawDataMetaChunks[rawindex].RawDataChunkInfo.Length == 0)
                        continue;

                    //go to where we are going to draw
                    Map.dev.Transform.World = bsp.PermutationInfo[x].mat;
                    Map.dev.SetStreamSource(0, bsp.Display.permvertexBuffer[rawindex], 0);
                    Map.dev.VertexFormat = ParsedModel.HaloVertex.FVF;
                    Map.dev.Indices = bsp.Display.permindexBuffer[rawindex];

                    for (int xx = 0; xx < bsp.BSPPermutationRawDataMetaChunks[rawindex].SubMeshInfo.Length; xx++)
                    {
                        if (fm == FillMode.Solid)
                        {
                            Map.dev.RenderState.CullMode = Cull.Clockwise;

                            ResetTextureStates();
                            int tempshade = bsp.BSPPermutationRawDataMetaChunks[rawindex].SubMeshInfo[xx].ShaderNumber;

                            // AlphaBlending
                            if (bsp.Shaders.Shader[tempshade].Alpha == ShaderInfo.AlphaType.AlphaBlend)
                            {
                                Map.dev.RenderState.AlphaBlendEnable = true;
                                Map.dev.RenderState.AlphaTestEnable = false;
                            }
                            else if (bsp.Shaders.Shader[tempshade].Alpha == ShaderInfo.AlphaType.AlphaTest)
                            {
                                Map.dev.RenderState.AlphaBlendEnable = false;
                                Map.dev.RenderState.AlphaTestEnable = true;
                            }
                            else
                            {
                                Map.dev.RenderState.AlphaBlendEnable = false;
                                Map.dev.RenderState.AlphaTestEnable = false;
                            }

                            // ChooseTexture
                            Map.dev.SetTexture(0, bsp.Shaders.Shader[tempshade].MainTexture);

                            Map.dev.TextureState[0].ColorOperation = TextureOperation.Modulate;
                            Map.dev.TextureState[0].ColorArgument1 = TextureArgument.TextureColor;
                            Map.dev.TextureState[0].ColorArgument2 = TextureArgument.Current;

                            if (bsp.SceneryLightMapTexture[x] != null && bsp.BSPPermutationRawDataMetaChunks[rawindex].LightMapUVs.Count != 0)
                            {
                                Map.dev.SetTexture(1, bsp.SceneryLightMapTexture[x]);
                                Map.dev.TextureState[1].ColorOperation = TextureOperation.Modulate;
                                Map.dev.TextureState[1].ColorArgument1 = TextureArgument.TextureColor;
                                Map.dev.TextureState[1].ColorArgument2 = TextureArgument.Current;
                                Map.dev.TextureState[1].TextureCoordinateIndex = 2;
                            }
                            else
                            {
                                Map.dev.TextureState[0].ColorOperation = TextureOperation.Disable;
                                // Map.dev.TextureState[0].ColorOperation = TextureOperation.Modulate2X;
                                // Map.dev.TextureState[0].ColorArgument1 = TextureArgument.TextureColor;
                                // Map.dev.TextureState[0].ColorArgument2 = TextureArgument.Current;
                            }

                            Map.dev.RenderState.FillMode = FillMode.Solid;

                            Material mm = new Material();
                            mm.Diffuse = mm.Ambient = System.Drawing.Color.White;
                            Map.dev.Material = mm;

                            Map.dev.DrawIndexedPrimitives(
                                 PrimitiveType.TriangleList,
                                 0,
                                 0,
                                 bsp.BSPPermutationRawDataMetaChunks[rawindex].VerticeCount,
                                 bsp.BSPPermutationRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceStart,
                                 bsp.BSPPermutationRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceCount / 3
                              );
                        }
                        else //wireframe
                        {
                            Material m = new Material();

                            if (bsp.PermutationInfo[x].IsSelected)
                                m.Diffuse = m.Ambient = System.Drawing.Color.Red;

                            else m.Diffuse = m.Ambient = System.Drawing.Color.Black;

                            Map.dev.Material = m;
                            Map.dev.RenderState.FillMode = FillMode.WireFrame;
                            Map.dev.DrawIndexedPrimitives(
                                PrimitiveType.TriangleList,
                                0,
                                0,
                                bsp.BSPPermutationRawDataMetaChunks[rawindex].VerticeCount,
                                bsp.BSPPermutationRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceStart,
                                bsp.BSPPermutationRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceCount / 3
                             );
                        }

                        //if item is selected and we did just draw textures
                        if (fm == FillMode.Solid && bsp.PermutationInfo[x].IsSelected)
                        {
                            Material m = new Material();
                            m.Diffuse = m.Ambient = System.Drawing.Color.Red;
                            Map.dev.Material = m;

                            Map.dev.RenderState.FillMode = FillMode.WireFrame;
                           // Map.dev.Transform.World =
                           //     bsp.BSPPermutationRawDataMetaChunks[rawindex].BoundingBox.mat;

                            //bsp.BSPPermutationRawDataMetaChunks[rawindex].Box.DrawSubset(0);

                            Map.dev.DrawIndexedPrimitives(
                                PrimitiveType.TriangleList,
                                0,
                                0,
                                bsp.BSPPermutationRawDataMetaChunks[rawindex].VerticeCount,
                                bsp.BSPPermutationRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceStart,
                                bsp.BSPPermutationRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceCount / 3
                             );
                        }
                    }
                }
                ResetTextureStates();
            }

            public static void DrawCluster(int rawindex, System.Drawing.Color outlinecolor,
                ref BSPModel bsp, bool Textured, bool WireFrame, ref Camera cam, int mapnum)
            {
                // BSPMainMesh
                if (bsp.BSPRawDataMetaChunks[rawindex].RawDataChunkInfo.Length == 0)
                    return;

                //   Vector3 tempv = new Vector3();
                //   tempv.X = bsp.Spawns.Spawn[x].X;
                //   tempv.Y = bsp.Spawns.Spawn[x].Y;
                //  tempv.Z = bsp.Spawns.Spawn[x].Z;
                //   if (!cam.SphereInFrustum(tempv, 10f)) { continue; }

                Map.dev.SetStreamSource(0, bsp.Display.vertexBuffer[rawindex], 0);
                Map.dev.VertexFormat = CustomVertex.PositionNormalTextured.Format;
                Map.dev.Indices = bsp.Display.indexBuffer[rawindex];
                for (int xx = 0; xx < bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo.Length; xx++)
                {
                    int tempshade = bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[xx].ShaderNumber;

                    // AlphaBlending
                    if (bsp.Shaders.Shader[tempshade].Alpha == ShaderInfo.AlphaType.AlphaBlend)
                    {
                        //  MessageBox.Show("test");
                        Map.dev.RenderState.AlphaBlendEnable = true;
                        Map.dev.RenderState.AlphaTestEnable = false;
                    }
                    else if (bsp.Shaders.Shader[tempshade].Alpha == ShaderInfo.AlphaType.AlphaTest)
                    {
                        // MessageBox.Show("test2");
                        Map.dev.RenderState.AlphaBlendEnable = false;
                        Map.dev.RenderState.AlphaTestEnable = true;
                    }
                    else
                    {
                        Map.dev.RenderState.AlphaBlendEnable = false;
                        Map.dev.RenderState.AlphaTestEnable = false;
                    }

                    // ChooseTexture
                    if (Textured == true)
                    {
                        // Map.dev.TextureState[0] = Map.dev.TextureState[2];
                        Map.dev.SetTexture(0, bsp.Shaders.Shader[tempshade].MainTexture);
                    }
                    else Map.dev.SetTexture(0, null);

                    // RenderToScene
                    PrimitiveType pt;
                    pt = PrimitiveType.TriangleList;
                    Map.dev.RenderState.FillMode = FillMode.Solid;
                    Material mm = new Material();
                    mm.Diffuse = System.Drawing.Color.White;
                    mm.Ambient = System.Drawing.Color.White;
                    Map.dev.Material = mm;
                    Map.dev.DrawIndexedPrimitives(
                        pt,
                        0,
                        0,
                        bsp.BSPRawDataMetaChunks[rawindex].VerticeCount,
                        bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceStart,
                        bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceCount / 3
                     );

                    // WireFrame
                    if (WireFrame == true)
                    {
                        Material m = new Material();
                        m.Diffuse = outlinecolor;
                        m.Ambient = outlinecolor;
                        Map.dev.Material = m;
                        Map.dev.RenderState.FillMode = FillMode.WireFrame;
                        Map.dev.DrawIndexedPrimitives(
                            pt,
                            0,
                            0,
                            bsp.BSPRawDataMetaChunks[rawindex].VerticeCount,
                            bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceStart,
                            bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceCount / 3);

                    }
                }
            }
        }
    }
}
