using System;
using System.Collections.Generic;
using System.Text;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.IO;
using System.Drawing;

namespace Visual_Scenery_Mover
{
    public class ParsedModel : IDisposable
    {
        public struct HaloVertex
        {
            public Vector3 Position, Normal;

            public int diffuse, specular;
            public float Tu0, Tv0;
            // public float Tu1, Tv1;
            public static readonly VertexFormats FVF =
                VertexFormats.PositionNormal | VertexFormats.Diffuse |
                VertexFormats.Specular | VertexFormats.Texture1;
        }

        public BoundingBoxContainer BoundingBox;
        public LODInfo LOD;
        public RawDataMetaChunk[] RawDataMetaChunks;
        public ShaderContainer Shaders;
        public hlmtContainer hlmt;
        public DisplayedInfo Display;
        public string name;
        public string PermutationString;
        public FrameHierarchy Frames;

        public ParsedModel()
        {
        }

        public ParsedModel(ref Meta meta, int mapnum)
        {
            BinaryReader BR = new BinaryReader(meta.MS);

            string[] temps = meta.name.Split('\\');
            name = temps[temps.Length - 1];
            BoundingBox = new BoundingBoxContainer(ref meta);

            BR.BaseStream.Position = 36;
            int tempc = BR.ReadInt32();
            int tempr = BR.ReadInt32() - meta.magic - meta.offset;
            RawDataMetaChunks = new RawDataMetaChunk[tempc];
            for (int x = 0; x < tempc; x++)
                RawDataMetaChunks[x] = new
                    RawDataMetaChunk(tempr + (x * 92), x, BoundingBox, ref meta);

            Shaders = new ShaderContainer(ref meta, mapnum);
            LOD = new LODInfo(ref meta, ref RawDataMetaChunks, mapnum);

            int temphlmt = -1;
            for (int x = 0; x < Map.mp[mapnum].tags.tagInfo.Length; x++)
            {
                if ("hlmt" == Map.mp[mapnum].tags.tagInfo[x].TagType && Map.mp[mapnum].tags.tagInfo[x].TagPath == meta.name)
                    temphlmt = x;
            }

            if (temphlmt != -1)
            {
                hlmt = new hlmtContainer(temphlmt, mapnum);
                if (hlmt.Permutations.Length > 0)
                    PermutationString = hlmt.Permutations[0].Name;
            }

            Display = DisplayedInfo.FindDisplayedPieces(4, this);

            Frames = new FrameHierarchy();
            Frames.GetFramesFromHalo2Model(ref meta, mapnum);
        }

        public float DecompressVertice(float input, float min, float max)
        {
            float percent = (input + 32768) / 65535;
            float result = (percent * (max - min)) + min;
            return result;
        }

        public static Vector3 DecompressNormal(int compressednormal)
        {
            int xx = compressednormal & 0x3ff;
            float xxx = Convert.ToSingle(xx) / Convert.ToSingle(0x3ff);
            int bitx = ((compressednormal >> 10) & 0x1);
            if (bitx == 1) { xxx = 1 - xxx; xxx = -xxx; }

            int yy = ((compressednormal >> 11) & 0x3ff);
            float yyy = Convert.ToSingle(yy) / Convert.ToSingle(0x3ff);
            int bity = ((compressednormal >> 21) & 0x1);
            if (bity == 1) { yyy = 1 - yyy; yyy = -yyy; }

            int zz = ((compressednormal >> 22) & 0x1ff);
            float zzz = Convert.ToSingle(zz) / Convert.ToSingle(0x1ff);
            int bitz = ((compressednormal >> 31) & 0x1);
            if (bitz == 1) { zzz = 1 - zzz; zzz = -zzz; }
            return new Vector3(xxx, yyy, zzz);
        }

        public static int CompressNormal(Vector3 normal)
        {
            bool negx = false;
            if (normal.X < 0) { negx = true; normal.X = -normal.X; }
            float bbb = normal.X * Convert.ToSingle(0x3ff);
            int b = Convert.ToInt32(bbb);
            if (negx) b = -b + 1 | 0x400;

            bool negy = false;
            if (normal.Y < 0) { negy = true; normal.Y = -normal.Y; }
            float bbb2 = normal.Y * Convert.ToSingle(0x3ff);
            int b2 = Convert.ToInt32(bbb2);
            if (negy) b2 = -b2 + 1 | 0x400;
            b2 <<= 11;

            bool negz = false;
            if (normal.Z < 0) { negz = true; normal.Z = -normal.Z; }
            float bbb3 = normal.Z * Convert.ToSingle(0x1ff);
            int b3 = Convert.ToInt32(bbb3);
            if (negz) b3 = -b3 + 1 | 0x200;
            b3 <<= 22;

            int result = b3 | b2 | b;

            return result;
        }

        public void Dispose()
        {
            this.BoundingBox = null;
            this.Display = null;
            this.hlmt = null;
            this.LOD = null;
            this.RawDataMetaChunks = null;
            this.Shaders = null;
            GC.SuppressFinalize(this);
        }

        public class BoundingBoxContainer
        {
            public float MinX = 0f;
            public float MaxX = 0f;
            public float MinY = 0f;
            public float MaxY = 0f;
            public float MinZ = 0f;
            public float MaxZ = 0f;
            public float MinU = 0f;
            public float MaxU = 0f;
            public float MinV = 0f;
            public float MaxV = 0f;

            public BoundingBoxContainer()
            {
            }

            public BoundingBoxContainer(ref Meta meta)
            {
                BinaryReader BR = new BinaryReader(meta.MS);
                BR.BaseStream.Position = 20;
                int tempc = BR.ReadInt32();
                int tempr = BR.ReadInt32() - meta.magic - meta.offset;
                BR.BaseStream.Position = tempr;
                MinX = BR.ReadSingle();
                MaxX = BR.ReadSingle();
                MinY = BR.ReadSingle();
                MaxY = BR.ReadSingle();
                MinZ = BR.ReadSingle();
                MaxZ = BR.ReadSingle();
                MinU = BR.ReadSingle();
                MaxU = BR.ReadSingle();
                MinV = BR.ReadSingle();
                MaxV = BR.ReadSingle();
            }
        }

        public class LODInfo
        {
            public LODPieceInfo[] Piece;
            public List<string> PermutationStrings = new List<string>();

            public LODInfo(ref Meta meta, ref RawDataMetaChunk[] rd, int mapnum)
            {
                BinaryReader BR = new BinaryReader(meta.MS);
                BR.BaseStream.Position = 28;
                int tempc = BR.ReadInt32();
                int tempr = BR.ReadInt32() - meta.magic - meta.offset;
                Piece = new LODPieceInfo[tempc];

                for (int x = 0; x < tempc; x++)
                {
                    Piece[x] = new LODPieceInfo();
                    BR.BaseStream.Position = tempr + (x * 16);
                    Piece[x].Name = Map.mp[mapnum].strings.stringsList[BR.ReadInt16()];
                    BR.BaseStream.Position = tempr + (x * 16) + 8;
                    int tempc2 = BR.ReadInt32();
                    int tempr2 = BR.ReadInt32() - meta.magic - meta.offset;
                    Piece[x].Permutation = new LODPieceInfo.LODPermutationInfo[tempc2];
                    for (int xx = 0; xx < tempc2; xx++)
                    {
                        BR.BaseStream.Position = tempr2 + (xx * 16);
                        Piece[x].Permutation[xx] = new LODPieceInfo.LODPermutationInfo();
                        Piece[x].Permutation[xx].name = Map.mp[mapnum].strings.stringsList[BR.ReadInt16()];
                        if (PermutationStrings.IndexOf(Piece[x].Permutation[xx].name) == -1) { PermutationStrings.Add(Piece[x].Permutation[xx].name); }
                        BR.ReadInt16();
                        Piece[x].Permutation[xx].pieceNumber = new int[5];
                        for (int w = 0; w < 5; w++)
                        {
                            Piece[x].Permutation[xx].pieceNumber[w] = BR.ReadInt16();
                            rd[Piece[x].Permutation[xx].pieceNumber[w]].piecename = Piece[x].Name;
                            rd[Piece[x].Permutation[xx].pieceNumber[w]].permutation = Piece[x].Permutation[xx].name;
                            rd[Piece[x].Permutation[xx].pieceNumber[w]].lod = w;
                        }
                    }
                }
            }

            public LODInfo.LODPieceInfo.LODPermutationInfo FindPermutationInfo(string piecename, string permutationname)
            {
                LODInfo.LODPieceInfo.LODPermutationInfo temp = null;
                for (int x = 0; x < Piece.Length; x++)
                {
                    if (Piece[x].Name == piecename)
                    {
                        for (int xx = 0; xx < Piece[x].Permutation.Length; xx++)
                        {
                            if (Piece[x].Permutation[xx].name == permutationname)
                            {
                                return Piece[x].Permutation[xx];
                            }
                        }
                        temp = Piece[x].Permutation[0];
                    }
                }
                return temp;
            }

            public class LODPieceInfo
            {
                public string Name;
                public LODPermutationInfo[] Permutation;
                public class LODPermutationInfo
                {
                    public string name;
                    public int[] pieceNumber;
                }
            }
        }

        public class PRTMRawDataMetaChunk : RawDataMetaChunk
        {
            public PRTMRawDataMetaChunk(int mapnum)
            {
                Meta meta = Map.mp[mapnum].meta;
                BinaryReader BR = new BinaryReader(meta.MS);

                BR.BaseStream.Position = 176;
                int tempc = BR.ReadInt32();
                int tempr = BR.ReadInt32() - Map.mp[mapnum].index.secondaryMagic - meta.offset;
                RawDataChunkInfo = new RawDataOffsetChunk[tempc];
                for (int x = 0; x < tempc; x++)
                {
                    RawDataChunkInfo[x] = new RawDataOffsetChunk();
                    BR.BaseStream.Position = tempr + (x * 16) + 6;
                    RawDataChunkInfo[x].ChunkSize = 56;
                    RawDataChunkInfo[x].Size = BR.ReadInt32();
                    if (RawDataChunkInfo[x].ChunkSize == 0) { RawDataChunkInfo[x].ChunkSize = RawDataChunkInfo[x].Size; }
                    RawDataChunkInfo[x].Offset = BR.ReadInt32();
                    RawDataChunkInfo[x].ChunkCount = RawDataChunkInfo[x].Size / RawDataChunkInfo[x].ChunkSize;
                }

                BR.BaseStream.Position = 136;
                VerticeCount = BR.ReadInt32();
                tempr = BR.ReadInt32() - Map.mp[mapnum].index.secondaryMagic - meta.offset;
                BR.BaseStream.Position = tempr;
                for (int x = 0; x < VerticeCount; x++)
                {
                    Vector3 vec = new Vector3();

                    vec.X = BR.ReadSingle();
                    vec.Y = BR.ReadSingle();
                    vec.Z = BR.ReadSingle();
                    Vertices.Add(vec);

                    Vector3 vec2 = new Vector3();
                    vec2.X = BR.ReadSingle();
                    vec2.Y = BR.ReadSingle();
                    vec2.Z = BR.ReadSingle();
                    Normals.Add(vec2);

                    Vector3 vec3 = new Vector3();
                    vec3.X = BR.ReadSingle();
                    vec3.Y = BR.ReadSingle();
                    vec3.Z = BR.ReadSingle();
                    Binormals.Add(vec3);

                    Vector3 vec4 = new Vector3();
                    vec4.X = BR.ReadSingle();
                    vec4.Y = BR.ReadSingle();
                    vec4.Z = BR.ReadSingle();
                    Tangents.Add(vec4);

                    Vector2 vec5 = new Vector2();
                    vec5.X = BR.ReadSingle();
                    vec5.Y = BR.ReadSingle();

                    UVs.Add(vec5);
                }

                BR.BaseStream.Position = 168;
                HeaderSize = BR.ReadInt32() + 8;

                BR.BaseStream.Position = 144;
                IndiceCount = BR.ReadInt32();
                tempr = BR.ReadInt32() - Map.mp[mapnum].index.secondaryMagic - meta.offset;
                BR.BaseStream.Position = tempr;
                FaceCount = IndiceCount / 3;

                this.Indices = new List<short>(); ;
                for (int x = 0; x < IndiceCount; x++)
                    Indices.Add((short)BR.ReadUInt16());

                SubMeshInfo = new ModelSubMeshInfo[1];
                for (int x = 0; x < 1; x++)
                {
                    SubMeshInfo[x] = new ModelSubMeshInfo();
                    BR.BaseStream.Position = HeaderSize + RawDataChunkInfo[0].Offset + (x * 72) + 4;
                    SubMeshInfo[x].ShaderNumber = 0;
                    SubMeshInfo[x].IndiceStart = 0;
                    SubMeshInfo[x].IndiceCount = IndiceCount;

                }
            }
        }

        public class RawDataMetaChunk
        {
            public RawDataOffsetChunk[] RawDataChunkInfo;
            public int VerticeCount;
            public int FaceCount;

            public bool IsSelected = false;

            public Mesh mesh;
            public Material RedMaterial = new Material();

            public byte type;
            public int IndiceCount;

            //public short[] Indices;
            public List<short> Indices = new List<short>();

            public List<Vector3> Vertices = new List<Vector3>();
            public List<Vector3> Normals = new List<Vector3>();
            public List<Vector3> Binormals = new List<Vector3>();
            public List<Vector3> Tangents = new List<Vector3>();

            public List<Vector2> UVs = new List<Vector2>();
            public List<BoneInfo> VerticeBones = new List<BoneInfo>();
            public List<int> BoneMap = new List<int>();
            public int HeaderSize;
            public ModelSubMeshInfo[] SubMeshInfo;
            public string piecename;
            public string permutation;
            public int lod;

            public RawDataMetaChunk()
            { }

            public RawDataMetaChunk(int offset, int chunknumber,
                BoundingBoxContainer bb, ref Meta meta)
            {
                RedMaterial.Diffuse = RedMaterial.Ambient = Color.Red;

                BinaryReader BR = new BinaryReader(meta.MS);

                BR.BaseStream.Position = offset + 4;
                VerticeCount = BR.ReadUInt16();
                FaceCount = BR.ReadUInt16();
                BR.BaseStream.Position = offset + 20;
                type = BR.ReadByte();
                BR.BaseStream.Position = offset + 68;
                HeaderSize = ((RawDataChunk)meta.raw.rawChunks[chunknumber]).size - BR.ReadInt32() - 4;
                int tempc = BR.ReadInt32();
                int tempr = BR.ReadInt32() - meta.magic - meta.offset;
                RawDataChunkInfo = new RawDataOffsetChunk[tempc];
                for (int x = 0; x < tempc; x++)
                {
                    RawDataChunkInfo[x] = new RawDataOffsetChunk();
                    BR.BaseStream.Position = tempr + (x * 16) + 6;
                    RawDataChunkInfo[x].ChunkSize = BR.ReadUInt16();
                    RawDataChunkInfo[x].Size = BR.ReadInt32();
                    if (RawDataChunkInfo[x].ChunkSize == 0) { RawDataChunkInfo[x].ChunkSize = RawDataChunkInfo[x].Size; }
                    RawDataChunkInfo[x].Offset = BR.ReadInt32();
                    RawDataChunkInfo[x].ChunkCount = RawDataChunkInfo[x].Size / RawDataChunkInfo[x].ChunkSize;
                }

                BR = new BinaryReader(((RawDataChunk)meta.raw.rawChunks[chunknumber]).MS);
                SubMeshInfo = new ModelSubMeshInfo[RawDataChunkInfo[0].ChunkCount];
                for (int x = 0; x < RawDataChunkInfo[0].ChunkCount; x++)
                {
                    SubMeshInfo[x] = new ModelSubMeshInfo();
                    BR.BaseStream.Position = HeaderSize + RawDataChunkInfo[0].Offset + (x * 72) + 4;
                    SubMeshInfo[x].ShaderNumber = BR.ReadUInt16();
                    SubMeshInfo[x].IndiceStart = BR.ReadUInt16();
                    SubMeshInfo[x].IndiceCount = BR.ReadUInt16();
                }

                BR.BaseStream.Position = 40;
                IndiceCount = BR.ReadUInt16();

                #region Determine_Raw_Data_Chunk_Numbers
                int indicechunk = 0;
                int verticechunk = 0;
                int uvchunk = 0;

                for (int x = 0; x < RawDataChunkInfo.Length; x++)
                {
                    if (RawDataChunkInfo[x].ChunkSize == 2)
                    {
                        indicechunk = x;
                        verticechunk = x + 2;
                        uvchunk = x + 3;
                        break;
                    }
                }

                int bonemapchunk = 0;
                BR.BaseStream.Position = 108;
                int tempbonemapcount = BR.ReadUInt16();
                if (tempbonemapcount > 0)
                {
                    for (int x = uvchunk + 1; x < RawDataChunkInfo.Length; x++)
                    {
                        if (RawDataChunkInfo[x].ChunkSize == 1)
                        {
                            bonemapchunk = x;
                            break;
                        }
                    }
                    BR.BaseStream.Position = HeaderSize + RawDataChunkInfo[bonemapchunk].Offset;
                    for (int x = 0; x < tempbonemapcount; x++)
                    {
                        BoneMap.Add(BR.ReadByte());
                    }
                }
                else
                {
                    BoneMap.Add(0);
                }
                #endregion

                RawDataChunkInfo[verticechunk].ChunkSize = RawDataChunkInfo[verticechunk].Size / VerticeCount;
                for (int x = 0; x < VerticeCount; x++)
                {
                    Vector3 vec = new Vector3();
                    BR.BaseStream.Position = HeaderSize + RawDataChunkInfo[verticechunk].Offset + (RawDataChunkInfo[verticechunk].ChunkSize * x);
                    vec.X = DecompressVertice(BR.ReadInt16(), bb.MinX, bb.MaxX);
                    vec.Y = DecompressVertice(BR.ReadInt16(), bb.MinY, bb.MaxY);
                    vec.Z = DecompressVertice(BR.ReadInt16(), bb.MinZ, bb.MaxZ);
                    Vertices.Add(vec);
                    //   if (tempbonemapcount == 0) { continue; }

                    switch (RawDataChunkInfo[verticechunk].ChunkSize)
                    {
                        case 6:
                            BoneInfo b = new BoneInfo();
                            b.BoneIndex.Add(0);
                            b.Weight.Add(1.0f);
                            VerticeBones.Add(b);

                            break;
                        case 8:
                            BoneInfo c = new BoneInfo();
                            c.BoneIndex.Add(BR.ReadByte());
                            c.Weight.Add(1.0f);

                            byte tempb = BR.ReadByte();
                            if (tempb == 0) { VerticeBones.Add(c); break; }
                            BoneInfo c2 = new BoneInfo();
                            c2.BoneIndex.Add(tempb);
                            c2.Weight.Add(1.0f);
                            VerticeBones.Add(c2);
                            c.Weight[0] = 1.0f;
                            VerticeBones.Add(c);
                            break;
                        case 12:
                            BoneInfo bbb = new BoneInfo();
                            bbb.BoneIndex.Add(BR.ReadByte());
                            bbb.Weight.Add(0.99f);
                            VerticeBones.Add(bbb);
                            break;
                    }
                }

                RawDataChunkInfo[uvchunk].ChunkSize = 4;
                for (int x = 0; x < VerticeCount; x++)
                {
                    Vector2 tempuv = new Vector2();
                    BR.BaseStream.Position = HeaderSize + RawDataChunkInfo[uvchunk].Offset + (RawDataChunkInfo[uvchunk].ChunkSize * x);
                    tempuv.X = DecompressVertice(BR.ReadInt16(), bb.MinU, bb.MaxU) % 1;
                    tempuv.Y = DecompressVertice(BR.ReadInt16(), bb.MinV, bb.MaxV) % 1;

                    // if (tempuv.X > 1) { tempuv.X = tempuv.X - 1; }
                    // else 
                    if (tempuv.X < 0) { tempuv.X = 1 - tempuv.X; }

                    // if (tempuv.Y > 1) { tempuv.Y = tempuv.Y - 1; }
                    // else 
                    if (tempuv.Y < 0) { tempuv.Y = 1 - tempuv.Y; }
                    UVs.Add(tempuv);
                }

                RawDataChunkInfo[uvchunk + 1].ChunkSize = 12;
                for (int x = 0; x < VerticeCount; x++)
                {
                    BR.BaseStream.Position = HeaderSize + RawDataChunkInfo[uvchunk + 1].Offset + (RawDataChunkInfo[uvchunk + 1].ChunkSize * x);

                    int dword = BR.ReadInt32();

                    Vector3 tempnormal = ParsedModel.DecompressNormal(dword);

                    int converto = ParsedModel.CompressNormal(tempnormal);
                    Vector3 tempnormal2 = ParsedModel.DecompressNormal(converto);
                    Normals.Add(tempnormal);
                }

                BR.BaseStream.Position = HeaderSize + RawDataChunkInfo[indicechunk].Offset;
                this.Indices = new List<short>();
                for (int x = 0; x < IndiceCount; x++)
                    Indices.Add((short)BR.ReadUInt16());
            }

            public float DecompressVertice(float input, float min, float max)
            {
                float percent = (input + 32768) / 65535;
                float result = (percent * (max - min)) + min;
                return result;
            }

            public float CompressVertice(float input, float min, float max)
            {
                float result = input - min;
                result = (result / (max - min));
                result = result * 65535;
                result = result - 32768;
                return result;
            }

            public class ModelSubMeshInfo
            {
                public int ShaderNumber;
                public int IndiceStart;
                public int IndiceCount;
                public Mesh mesh;
            }

            public class RawDataOffsetChunk
            {
                public int Offset;
                public int Size;
                public int ChunkSize;
                public int ChunkCount;
            }

            public class BoneInfo
            {
                public List<int> BoneIndex = new List<int>();
                public List<float> Weight = new List<float>();
            }
        }

        public class ShaderContainer
        {
            public ShaderInfo[] Shader;
            public int count = 0;

            public ShaderContainer()
            {
            }

            public ShaderContainer(ref Meta meta, int mapnum)
            {
                BinaryReader BR = new BinaryReader(meta.MS);
                BR.BaseStream.Position = 96;
                int tempc = BR.ReadInt32();
                int tempr = BR.ReadInt32() - meta.magic - meta.offset;
                Shader = new ShaderInfo[tempc];

                count = tempc;
                for (int x = 0; x < tempc; x++)
                {
                    BR.BaseStream.Position = tempr + (x * 32) + 12;
                    int tempint = Meta.FindMetaByID(BR.ReadInt32(), Map.mp[mapnum].tags.tagInfo);

                    Shader[x] = new ShaderInfo(tempint, mapnum);
                }
            }
        }

        public class DisplayedInfo
        {
            public List<int> Chunk = new List<int>();
            public int[] ShaderIndex;
            public VertexBuffer[] vertexBuffer = null;
            public IndexBuffer[] indexBuffer = null;
            public Mesh[] meshes;

            public static DisplayedInfo FindDisplayedPieces(int lodlevel, ParsedModel pm)
            {
                DisplayedInfo di = new DisplayedInfo();
                if (pm.hlmt == null || pm.hlmt.Permutations.Length == 0)
                {
                    for (int x = 0; x < pm.LOD.Piece.Length; x++)
                    {
                        for (int y = 0; y < pm.LOD.Piece[x].Permutation.Length; y++)
                        {
                            if (pm.LOD.Piece[x].Permutation[y].name == pm.LOD.PermutationStrings[0])
                            {
                                if (di.Chunk.IndexOf(pm.LOD.Piece[x].Permutation[y].pieceNumber[lodlevel]) == -1)
                                {
                                    di.Chunk.Add(pm.LOD.Piece[x].Permutation[y].pieceNumber[lodlevel]);
                                }
                            }
                        }
                    }
                    goto shadeshit;
                }

                int tempid = pm.hlmt.FindPermutation(pm.PermutationString);
                ParsedModel.LODInfo.LODPieceInfo.LODPermutationInfo lodinfo = new ParsedModel.LODInfo.LODPieceInfo.LODPermutationInfo();
                for (int x = 0; x < pm.hlmt.Permutations[tempid].Piece.Length; x++)
                {
                    string name1 = pm.hlmt.Permutations[tempid].Piece[x].PieceName;
                    //                    if (pm.hlmt.Permutations[tempid].Piece[x].Permutation.Length
                    for (int xx = 0; xx < pm.hlmt.Permutations[tempid].Piece[x].Permutation.Length; xx++)
                    {
                        string name2 = pm.hlmt.Permutations[tempid].Piece[x].Permutation[xx].PermutationNameX;
                        lodinfo = pm.LOD.FindPermutationInfo(name1, name2);

                        if (di.Chunk.IndexOf(lodinfo.pieceNumber[lodlevel]) == -1)
                        {
                            di.Chunk.Add(lodinfo.pieceNumber[lodlevel]);
                        }
                    }
                }

            shadeshit:
                int[] tempshade = new int[200];
                int tempshadecount = 1;
                for (int x = 0; x < di.Chunk.Count; x++)
                {
                    for (int y = 0; y < pm.RawDataMetaChunks[di.Chunk[x]].SubMeshInfo.Length; y++)
                    {
                        int r = pm.RawDataMetaChunks[di.Chunk[x]].SubMeshInfo[y].ShaderNumber;
                        if (Array.IndexOf(tempshade, r) == -1)
                        {
                            tempshade[tempshadecount] = r;
                            tempshadecount++;
                        }
                    }
                }
                di.ShaderIndex = new int[tempshadecount];

                Array.Copy(tempshade, 0, di.ShaderIndex, 0, tempshadecount);
                return di;
            }

            public static void LoadDirectXTexturesAndBuffers(ref ParsedModel pm)
            {
                if (pm == null) return;
                DisplayedInfo.CreateVertexBuffers(ref pm);
                DisplayedInfo.CreateIndexBuffers(ref pm);
                DisplayedInfo.LoadShaderTextures(ref pm);
                DisplayedInfo.MakeMeshes(ref pm);
            }

            public static void MakeMeshes(ref ParsedModel pm)
            {
                pm.Display.meshes = new Mesh[pm.Display.Chunk.Count];
                for (int t = 0; t < pm.Display.Chunk.Count; t++)
                {
                    int rawindex = pm.Display.Chunk[t];
                    pm.Display.meshes[t] = MakeMesh(ref pm, rawindex);
                }
            }

            public static Mesh MakeMesh(List<Vector3> vertices, List<short> Indices, List<Vector2> uvs)
            {
                Mesh m = new Mesh(Indices.Count / 3, vertices.Count, MeshFlags.Managed,
                    HaloVertex.FVF, Map.dev);

                HaloVertex[] tempv = new HaloVertex[vertices.Count];
                for (int x = 0; x < vertices.Count; x++)
                {
                    tempv[x].Position = vertices[x];
                    tempv[x].Tu0 = uvs[x].X;
                    tempv[x].Tv0 = uvs[x].Y;
                }
                VertexBuffer vb = m.VertexBuffer;
                vb.SetData(tempv, 0, LockFlags.None);
                vb.Unlock();

                IndexBuffer ibx = m.IndexBuffer;
                ibx.SetData(Indices.ToArray(), 0, LockFlags.None);

                return m;
            }

            public static Mesh MakeMesh(ref ParsedModel pm, int chunknumber)
            {
                List<short> indices = new List<short>();
                int[] facecount = new int[pm.RawDataMetaChunks[chunknumber].SubMeshInfo.Length];
                int[] facestart = new int[pm.RawDataMetaChunks[chunknumber].SubMeshInfo.Length];
                if (pm.RawDataMetaChunks[chunknumber].FaceCount * 3 !=
                    pm.RawDataMetaChunks[chunknumber].Indices.Count)
                {
                    int tempc = 0;
                    for (int x = 0; x < pm.RawDataMetaChunks[chunknumber].SubMeshInfo.Length; x++)
                    {
                        indices = DecompressIndices(pm.RawDataMetaChunks[chunknumber].
                            Indices, pm.RawDataMetaChunks[chunknumber].SubMeshInfo[x].IndiceStart,
                            pm.RawDataMetaChunks[chunknumber].SubMeshInfo[x].IndiceCount);

                        //Array.ConstrainedCopy((tempindices, 0, tempindicesx, tempc, tempindices.Count);

                        facecount[x] = indices.Count / 3;
                        facestart[x] = tempc;
                        tempc += indices.Count;
                    }
                    //indices = new List<short>();
                    //Array.Copy(tempindicesx, indices, tempindicesx.Length);
                }
                else
                {
                    indices = pm.RawDataMetaChunks[chunknumber].Indices;
                    for (int x = 0; x < pm.RawDataMetaChunks[chunknumber].SubMeshInfo.Length; x++)
                    {
                        facecount[x] = pm.RawDataMetaChunks[chunknumber].SubMeshInfo[x].IndiceCount / 3;
                        facestart[x] = pm.RawDataMetaChunks[chunknumber].SubMeshInfo[x].IndiceStart / 3;
                    }
                }

                Mesh m = new Mesh(indices.Count / 3, pm.RawDataMetaChunks[chunknumber].Vertices.Count,
                    MeshFlags.Managed, HaloVertex.FVF, Map.dev);

                VertexBuffer vb = m.VertexBuffer;
                HaloVertex[] tempv = new HaloVertex[pm.RawDataMetaChunks[chunknumber].Vertices.Count];
                for (int x = 0; x < pm.RawDataMetaChunks[chunknumber].Vertices.Count; x++)
                {
                    tempv[x].Position = pm.RawDataMetaChunks[chunknumber].Vertices[x];
                    tempv[x].Normal = pm.RawDataMetaChunks[chunknumber].Normals[x];
                    tempv[x].Tu0 = pm.RawDataMetaChunks[chunknumber].UVs[x].X;
                    tempv[x].Tv0 = pm.RawDataMetaChunks[chunknumber].UVs[x].Y;

                }
                vb.SetData(tempv, 0, LockFlags.None);
                vb.Unlock();

                IndexBuffer ibx = m.IndexBuffer;
                ibx.SetData(indices.ToArray(), 0, LockFlags.None);

                return m;
            }

            public static List<short> DecompressIndices(List<short> indices, int start, int count)
            {
                bool dir = false;
                short tempx, tempy, tempz;
                List<short> shorts = new List<short>();
                int m = start, s = 0;
                do
                {
                    tempx = indices[m];
                    tempy = indices[m + 1];
                    tempz = indices[m + 2];

                    if (tempx != tempy && tempx != tempz && tempy != tempz)
                    {
                        if (dir == false)
                        {
                            shorts.Add(tempx);
                            shorts.Add(tempy);
                            shorts.Add(tempz);
                            s += 3;

                            dir = true;
                        }
                        else
                        {
                            shorts.Add(tempx);
                            shorts.Add(tempz);
                            shorts.Add(tempy);
                            s += 3;
                            dir = false;
                        }
                        m += 1;
                    }
                    else
                    {
                        if (dir == true) dir = false;
                        else dir = true; 

                        m += 1;
                    }
                }
                while (m < start + count - 2);
                return shorts;
            }

            private static void LoadShaderTextures(ref ParsedModel pm)
            {
                for (int x = 0; x < pm.Display.ShaderIndex.Length; x++)
                    pm.Shaders.Shader[pm.Display.ShaderIndex[x]].MakeTextures();
            }

            private static void CreateIndexBuffers(ref ParsedModel pm)
            {
                // Pool pool=;
                //  if (device.DeviceCaps..
                pm.Display.indexBuffer = new IndexBuffer[pm.RawDataMetaChunks.Length];
                for (int rawindex = 0; rawindex < pm.Display.Chunk.Count; rawindex++)
                {
                    int x = pm.Display.Chunk[rawindex];
                    pm.Display.indexBuffer[x] = new IndexBuffer(typeof(short), pm.RawDataMetaChunks[x].IndiceCount, Map.dev,Usage.WriteOnly, Pool.Managed);
                    IndexBuffer ib = (IndexBuffer)pm.Display.indexBuffer[x];
                    ib.SetData(pm.RawDataMetaChunks[x].Indices.ToArray(), 0, LockFlags.None);
                    ib.Unlock();
                }
            }

            private static void CreateVertexBuffers(ref ParsedModel pm)
            {
                pm.Display.vertexBuffer = new VertexBuffer[pm.RawDataMetaChunks.Length];
                for (int x = 0; x < pm.Display.Chunk.Count; x++)
                {
                    int rawindex = pm.Display.Chunk[x];

                    pm.Display.vertexBuffer[rawindex] = new VertexBuffer(typeof(HaloVertex), pm.RawDataMetaChunks[rawindex].VerticeCount, Map.dev,Usage.WriteOnly, HaloVertex.FVF, Pool.Managed);
                    HaloVertex[] verts = (HaloVertex[])pm.Display.vertexBuffer[rawindex].Lock(0, 0); // Lock the buffer (which will return our structs)
                    for (int i = 0; i < pm.RawDataMetaChunks[rawindex].VerticeCount; i++)
                    {
                        verts[i].Position = new Vector3(pm.RawDataMetaChunks[rawindex].Vertices[i].X,
                            pm.RawDataMetaChunks[rawindex].Vertices[i].Y,
                            pm.RawDataMetaChunks[rawindex].Vertices[i].Z);

                        verts[i].Tu0 = pm.RawDataMetaChunks[rawindex].UVs[i].X;
                        verts[i].Tv0 = pm.RawDataMetaChunks[rawindex].UVs[i].Y;
                        verts[i].Normal = new Vector3(pm.RawDataMetaChunks[rawindex].Vertices[i].X,
                            pm.RawDataMetaChunks[rawindex].Vertices[i].Y,
                            pm.RawDataMetaChunks[rawindex].Vertices[i].Z);

                        verts[i].specular = 1;
                        verts[i].diffuse = 1;

                    }
                    // Unlock (and copy) the data
                    pm.Display.vertexBuffer[rawindex].Unlock();
                }
            }

            public static void SetAlphaBlending(ShaderInfo.AlphaType alphatype)
            {
                if (alphatype == ShaderInfo.AlphaType.AlphaBlend)
                {
                    Map.dev.RenderState.AlphaBlendEnable = true;
                    Map.dev.RenderState.AlphaTestEnable = false;
                }
                else if (alphatype == ShaderInfo.AlphaType.AlphaTest)
                {
                    Map.dev.RenderState.AlphaBlendEnable = false;
                    Map.dev.RenderState.AlphaTestEnable = true;
                }
                else
                {
                    Map.dev.RenderState.AlphaBlendEnable = false;
                    Map.dev.RenderState.AlphaTestEnable = false;
                }
            }

            public static void Draw(ParsedModel pm, int mapnum)
            {
                for (int x = 0; x < pm.Display.Chunk.Count; x++)
                {
                    int rawindex = pm.Display.Chunk[x];

                    Map.dev.SetStreamSource(0, pm.Display.vertexBuffer[rawindex], 0);
                    Map.dev.VertexFormat = HaloVertex.FVF;
                    Map.dev.Indices = pm.Display.indexBuffer[rawindex];

                    for (int y = 0; y < pm.RawDataMetaChunks[rawindex].SubMeshInfo.Length; y++)
                    {
                        int tempshade = pm.RawDataMetaChunks[rawindex].SubMeshInfo[y].ShaderNumber;
                        SetAlphaBlending(pm.Shaders.Shader[tempshade].Alpha);

                        Map.dev.SetTexture(0, pm.Shaders.Shader[tempshade].MainTexture);
                        Map.dev.TextureState[0].ColorOperation = TextureOperation.SelectArg1;
                        Map.dev.TextureState[0].ColorArgument1 = TextureArgument.TextureColor;
                        Map.dev.TextureState[0].ColorArgument2 = TextureArgument.Current;
                        Map.dev.TextureState[0].AlphaOperation = TextureOperation.Modulate;
                        Map.dev.TextureState[0].AlphaArgument1 = TextureArgument.TextureColor;
                        Map.dev.RenderState.FillMode = FillMode.Solid;

                        //MessageBox.Show((pm.RawDataMetaChunks[rawindex].FaceCount * 3).ToString() + " :: " +
                        //     pm.RawDataMetaChunks[rawindex].IndiceCount.ToString());

                        if (pm.RawDataMetaChunks[rawindex].FaceCount * 3 != pm.RawDataMetaChunks[rawindex].IndiceCount)
                        {
                            Map.dev.DrawIndexedPrimitives(
                            PrimitiveType.TriangleStrip,
                            0,
                            0,
                            pm.RawDataMetaChunks[rawindex].VerticeCount,
                            pm.RawDataMetaChunks[rawindex].SubMeshInfo[y].IndiceStart,
                            pm.RawDataMetaChunks[rawindex].SubMeshInfo[y].IndiceCount-2);
                            //MessageBox.Show(pm.RawDataMetaChunks[rawindex].SubMeshInfo[y].IndiceCount.ToString());
                        }
                        else
                        {
                            Map.dev.DrawIndexedPrimitives(
                            PrimitiveType.TriangleList,
                            0,
                            0,
                            pm.RawDataMetaChunks[rawindex].VerticeCount,
                            pm.RawDataMetaChunks[rawindex].SubMeshInfo[y].IndiceStart,
                            pm.RawDataMetaChunks[rawindex].SubMeshInfo[y].IndiceCount / 3);
                        }
                    }
                }
            }

            public static void DrawMeshes(ParsedModel pm, bool textured)
            {
                for (int x = 0; x < pm.Display.Chunk.Count; x++)
                {
                    int rawindex = pm.Display.Chunk[x];
                    VertexBuffer vb = pm.Display.meshes[x].VertexBuffer;
                    IndexBuffer ib = pm.Display.meshes[x].IndexBuffer;

                    Map.dev.SetStreamSource(0, vb, 0);
                    Map.dev.VertexFormat = HaloVertex.FVF;
                    Map.dev.Indices = ib;

                    if (textured) Map.dev.RenderState.FillMode = FillMode.Solid;
                    else Map.dev.RenderState.FillMode = FillMode.WireFrame;

                    for (int y = 0; y < pm.RawDataMetaChunks[rawindex].SubMeshInfo.Length; y++)
                    {
                        int tempshade = pm.RawDataMetaChunks[rawindex].SubMeshInfo[y].ShaderNumber;
                        SetAlphaBlending(pm.Shaders.Shader[tempshade].Alpha);

                        if (textured)
                        {
                            Map.dev.SetTexture(0, pm.Shaders.Shader[tempshade].MainTexture);
                            Map.dev.TextureState[0].ColorOperation = TextureOperation.Modulate;
                            Map.dev.TextureState[0].ColorArgument1 = TextureArgument.TextureColor;
                            Map.dev.TextureState[0].ColorArgument2 = TextureArgument.Current;
                        }
                        else //wireframe
                        {
                            Map.dev.SetTexture(0, null);
                        }
                        pm.Display.meshes[x].DrawSubset(y);
                    }
                }
            }
        }
    }
}
